'''Unit tests for `substrings` package'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import pytest

from polymerist.genutils.textual.substrings import unique_string, shortest_repeating_substring, repeat_string_to_length


@pytest.mark.parametrize('string', ['Lorem', 'ipsum', 'dolor', 'sit', 'amet', 'consectetur', 'adipiscing', 'elit'])
def test_unique_str_unordered(string : str) -> None:
    '''Test that unique characters are coorectly identified WITHOUT respect to order'''
    assert set(unique_string(string, preserve_order=False)) == set(string)

@pytest.mark.parametrize('string, expected_output',
    [
        ('aaaaa', 'a'),
        ('BABAB', 'BA'),
        ('balaclava', 'balcv'),
        ('catamaran', 'catmrn'),
        ('unique', 'uniqe'), # self-reference makes everything better :P
        ('singular', 'singular'), # test string with aready-unique characters are unaffected
    ]
)
def test_unique_str_ordered(string : str, expected_output : str) -> None:
    '''Test that unique characters are coorectly identified WITH respect to order'''
    assert unique_string(string, preserve_order=True) == expected_output
    
    
@pytest.mark.parametrize('string, expected_output',
    [
        ('aaaaa', 'a'),
        ('booboo', 'boo'),
        ('piripiri', 'piri'),
        ('ababab', 'ab'),
        ('bcbabcbabcba', 'bcba'),
        # sequences which do not repeat a whole-number of times
        ('no repeats', 'no repeats'),
        ('ababa', 'ababa'), 
        ('bonobo', 'bonobo'),
    ]
)
def test_shortest_repeating_substring(string : str, expected_output : str) -> None:
    '''Test that minimal repeating substrings are correctly identified'''
    assert shortest_repeating_substring(string) == expected_output
    
    
@pytest.mark.parametrize('string, target_length, joiner, expected_output',
    [
        ('BACA', 10, '', 'BACABACABA'), # expected "standard" use case
        ('BACA', 1, '', 'B'),           # test case where target length is shorter than the whole string
        ('BACA', 0, '', ''),            # test that no repeats yields the empty string
        ('BACA', 4, '', 'BACA'),             # test precisely one repeat without joins
        ('BACA', 10, '|', 'BACA|BACA|BA'),   # test joiners
        ('BACA', 4, '|', 'BACA'),            # test no joiners are added when exactly one string repeat occurs
        ('BACA', 12, '|', 'BACA|BACA|BACA'), # test no extraneous joiners are included for purely-whole number of repeats
        ('CAT', 5, '', 'CATCA'), # test with triads (and different base string)
        pytest.param(''   ,   7, '', None, marks=pytest.mark.xfail(raises=ValueError, reason='Empty string can\'t be repeated into nonempty string', strict=True)),
        pytest.param('CAT', 4.2, '', None, marks=pytest.mark.xfail(raises=TypeError , reason='Non-integer string length doesn\'t make sense', strict=True)),
        pytest.param('CAT',  -1, '', None, marks=pytest.mark.xfail(raises=IndexError, reason='Can\'t have string with fewer than 0 characters', strict=True)),
    ]
)
def test_repeat_string_to_length(string : str, target_length : int, joiner : str, expected_output : str) -> None:
    '''Test that string repetition to a given length returns the expected string WITH joingin characters present'''
    assert repeat_string_to_length(string, target_length=target_length, joiner=joiner) == expected_output


'''For conversion of RDKit Mols back and forth between different format encodings - often imbues a desired side effect (such as 2D-projection)'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from abc import ABC, abstractmethod
from rdkit import Chem
from rdkit.Chem.rdchem import Mol

from ..genutils.decorators.classmod import register_subclasses, register_abstract_class_attrs

from .labeling.bijection import bijective_atom_id_iter
from .rdprops import copy_rd_props


@register_subclasses(key_attr='TAG')
@register_abstract_class_attrs('TAG')
class RDConverter(ABC): # TODO : add some optional sanitization measures to ensure valid output and bijection
    '''For converting an existing RDKit Molecule to and from a particular format to gain new properties'''
    @abstractmethod
    def _convert(self, rdmol : Mol) -> Mol:
        '''Implement conversion mechanism here'''
        pass

    def convert(self, rdmol : Mol, sanitize : bool=True) -> Mol:
        '''Tranform RDKit Mol using the selected method'''
        newmol = self._convert(rdmol)
        if sanitize:
            Chem.SanitizeMol(newmol)

        copy_rd_props(rdmol, newmol) # transfer properties at the mol- and atom-level to preserve consistency
        for atom_id_1, atom_id_2 in bijective_atom_id_iter(rdmol, newmol):
            orig_atom = rdmol.GetAtomWithIdx(atom_id_1)
            new_atom = newmol.GetAtomWithIdx(atom_id_2)

            copy_rd_props(orig_atom, new_atom)

        return newmol

class SMARTSConverter(RDConverter, TAG='SMARTS'):
    def _convert(self, rdmol : Mol) -> Mol:
        return Chem.MolFromSmarts(Chem.MolToSmarts(rdmol))

class SMILESConverter(RDConverter, TAG='SMILES'):
    def _convert(self, rdmol : Mol) -> Mol:
        return Chem.MolFromSmiles(Chem.MolToSmiles(rdmol), sanitize=False)
    
class CXSMARTSConverter(RDConverter, TAG='CXSMARTS'):
    '''Similar to SMARTSConverter but preserves the 3D structure'''
    def _convert(self, rdmol : Mol) -> Mol:
        return Chem.MolFromSmarts(Chem.MolToCXSmarts(rdmol))

class CXSMILESConverter(RDConverter, TAG='CXSMILES'):
    '''Similar to SMILESConverter but preserves the 3D structure'''
    def _convert(self, rdmol : Mol) -> Mol:
        return Chem.MolFromSmiles(Chem.MolToCXSmiles(rdmol), sanitize=False)

class InChIConverter(RDConverter, TAG='InChI'):
    # TOSELF : this does not preserve atom map num ordering (how to incorporate AuxInfo?)
    def _convert(self, rdmol : Mol) -> Mol:
        return Chem.AddHs(Chem.MolFromInchi(Chem.MolToInchi(rdmol), removeHs=False, sanitize=False))
    
class JSONConverter(RDConverter, TAG='JSON'):
    def _convert(self, rdmol : Mol) -> Mol:
        return Chem.rdMolInterchange.JSONToMols(Chem.MolToJSON(rdmol))[0]



'''Implementations of various number systems and representations of numbers'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from .commonbases import FixedRadixNumberSystem, FactorialNumberSystem

# initialization of some common bases
Factoradic = FactorialNumberSystem()

COMMON_BASES : dict[str, int] = {
    'binary'      : 2,
    'ternary'     : 3,
    'seximal'     : 6,
    'octal'       : 8,
    'decimal'     : 10,
    'duodecimal'  : 12,
    'hexadecimal' : 16,
}
for base_name, base in COMMON_BASES.items():
    base_sys = FixedRadixNumberSystem(base)
    globals()[base_name.capitalize()]     = base_sys # systematic name
    globals()[f'base{base}'.capitalize()] = base_sys # common name

    globals()[f'nega{base_name}'.capitalize()] = FixedRadixNumberSystem(-base) # also include negative base for kicks


'''For representing rational numbers, and more general ratios'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from dataclasses import dataclass
from typing import Any, Callable, ClassVar, TypeVar
N = TypeVar('N')

from numbers import Number
from math import gcd


# HELPER FUNCTIONS
def sgnmag(num : N) -> tuple[bool, N]:
    '''Returns the sign and magnitude of a numeric-like value'''
    return num < 0, abs(num)


# RATIO CLASSES
@dataclass(repr=False)
class Ratio:
    '''For representing fractional ratios between two objects'''
    num   : Any
    denom : Any

    # REPRESENTATION
    def __repr__(self) -> str:
        return f'{self.num}/{self.denom}'
    
    def to_latex(self) -> str:
        '''Return latex-compatible string which represent fraction'''
        return rf'\frac{{{self.num}}}{{{self.denom}}}'

    # RELATIONS
    @property
    def reciprocal(self) -> 'Ratio':
        '''Return the reciprocal of a ration'''
        return self.__class__(self.denom, self.num)

@dataclass(repr=False)
class Rational(Ratio):
    '''For representing ratios of integers'''
    num   : int
    denom : int

    # REDUCTION
    autoreduce : ClassVar[bool]=False
    
    def __post_init__(self) -> None:
        if self.__class__.autoreduce:
            self.reduce()

    def reduce(self) -> None:
        '''Reduce numerator and denominator by greatest common factor'''
        _gcd = gcd(self.num, self.denom)
        self.num=int(self.num / _gcd)
        self.denom=int(self.denom / _gcd)
    simplify = reduce # alias for convenience

    @property
    def reduced(self) -> 'Rational':
        '''Return reduced Rational equivalent to the current rational (does not modify in-place)'''
        new_rat = self.__class__(self.num, self.denom)
        new_rat.reduce()

        return new_rat
    simplifed = reduced # alias for convenience
    
    def as_proper(self) -> tuple[int, 'Rational']:
        '''Returns the integer and proper fractional component of a ratio'''
        integ, remain = divmod(self.num, self.denom)
        return integ, self.__class__(remain, self.denom)
    
    # ARITHMETIC
    def __add__(self, other : 'Rational') -> 'Rational':
        '''Sum of two Rationals'''
        return self.__class__(
            num=(self.num * other.denom) + (self.denom * other.num),
            denom=(self.denom * other.denom)
        )
    
    def __sub__(self, other : 'Rational') -> 'Rational':
        '''Difference of two Rationals'''
        return self.__class__(
            num=(self.num * other.denom) - (self.denom * other.num),
            denom=(self.denom * other.denom)
        )

    def __mul__(self, other : 'Rational') -> 'Rational':
        '''Product of two Rationals'''
        return self.__class__(
            num=self.num * other.num,
            denom=self.denom * other.denom
        )

    def __div__(self, other : 'Rational') -> 'Rational':
        '''Quotient of two Rationals'''
        return self.__class__(
            num=self.num * other.denom,
            denom=self.denom * other.num
        )
    
    def __pow__(self, power : float) -> 'Rational':
        '''Exponentiates a ratio'''
        return self.__class__(
            num=self.num**power,
            denom=self.denom**power
        )


'''Tools for simplifying logging from multiple sources'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
from logging import Logger
from traceback import format_exception

from types import ModuleType
from typing import Iterable, Optional, Union

from pathlib import Path
from datetime import datetime

from .timestamps import Timestamp, TIMESTAMP_LOG
from ..importutils.pkgiter import iter_submodules


# DATE AND TIME FORMATTING
LOG_FORMATTER = logging.Formatter('%(asctime)s.%(msecs)03d [%(levelname)-8s:%(module)16s:line %(lineno)-4d] - %(message)s', datefmt=TIMESTAMP_LOG) # TODO : make this more generic

def get_logger_registry() -> dict[str, Logger]:
    '''Gets all registered Loggers by module'''
    return logging.root.manager.loggerDict 

def get_active_loggers() -> list[Logger]:
    return [
        possible_logger
            for possible_logger in get_logger_registry().values()
                if isinstance(possible_logger, Logger) # omits PlaceHolder objects
    ]

def submodule_loggers(module : ModuleType, recursive : bool=True, blacklist : Optional[Iterable[str]]=None, sparse : bool=True) -> dict[str, Optional[logging.Logger]]:
    '''
    Produce a dict of any Logger objects present in each submodule. Can optionally generate recursively and blacklist certain modules
    
    Parameters
    ----------
    module : ModuleType
        The "root" module to begin importing from
        Represented in the Node object returned by this function
    recursive : bool, default=True
        Whether or not to recursively import modules from subpackages and add them to the tree
    blacklist : list[str] (optional), default None
        List of module names to exclude from tree building
        If provided, will exclude any modules whose names occur in this list
    sparse : bool, default=True
        Whether to only include modules which have a Logger defined (i.e. exclude all NoneType entries from returned dict)

    Returns
    -------
    logger_registry : dict[str, Optional[logging.Logger]]
        A dict keyed by module name whose values are the corresponding Logger bound to that module
    '''
    logger_registry = {}
    for module in iter_submodules(module, recursive=recursive, blacklist=blacklist):
        full_module_name = module.__name__
        module_logger = get_logger_registry().get(full_module_name, None)
        if isinstance(module, logging.PlaceHolder): 
            continue # exclude dummy Placeholder loggers

        if not (sparse and (module_logger is None)):
            logger_registry[full_module_name] = module_logger

    return logger_registry

# FILE-STREAM HANDLING CLASSES
class MultiStreamFileHandler(logging.FileHandler):
    '''Class to simplify logging file I/O given multiple logger sources providing logging input
    Automatically reports process completion & runtime if process is successful, or detailed error traceback otherwise
    
    Can spawn child processes to have multiple nested levels of logging to many partitioned output files for layered processes'''
    def __init__(self, filename : Union[str, Path], mode : str='a', encoding : Optional[str]=None, delay : bool=False, errors : Optional[str]=None, # FileHandler base args
                 loggers : Optional[Union[str, Logger, Iterable[Logger]]]='ALL', formatter : logging.Formatter=LOG_FORMATTER, proc_name : str='Process') -> None:       # args specific to this class
        super().__init__(filename, mode, encoding, delay, errors)

        self.proc_name : str = proc_name
        self.id : int = self.__hash__()  # generate unique ID number for tracking children
        
        self.setFormatter(formatter)
        self.personal_logger : Logger = logging.getLogger(str(self.id)) # create unique logger for internal error logging
        self.personal_logger.addHandler(self)

        self.parent : Optional[MultiStreamFileHandler] = None # to track whether the current process is the child of another process
        self.children : dict[int, MultiStreamFileHandler] = {} # keep track of child process; purely for debug (under normal circumstances, children unregister themselves once their process is complete)

        if loggers is None:
            return
        
        if isinstance(loggers, str):
            if loggers.lower() == 'all':
                loggers = get_active_loggers() # done inside of init to allow updating at runtime (rather than compile time)
            else:
                return
    
        self._loggers = []
        if isinstance(loggers, Logger): # only reachable if loggers is explicitly passed
            self.register_logger(loggers) # handle the singleton logger case
        else:
            self.register_loggers(*loggers) # handle a list of loggers

    def register_logger(self, logger : Logger) -> None:
        '''Add an individual Logger to the File Stream'''
        logger.addHandler(self)
        self._loggers.append(logger)

    def unregister_logger(self, logger : Logger) -> None:
        '''Remove an individual Logger from the collection of linked Loggers'''
        logger.removeHandler(self)
        logger_idx = self._loggers.index(logger)
        self._loggers.pop(logger_idx)
    
    def register_loggers(self, *loggers : list[Logger]) -> None:
        '''Record a new Logger and add the File handler to it - enables support for multiple Logger streams to a single file'''
        for logger in loggers:
            self.register_logger(logger)

    def unregister_loggers(self) -> None:
        '''Purge all currently registered Loggers'''
        for logger in self._loggers: # TOSELF : can't just call individual unregister_logger() method, since pop will change list size while iterating (causes some Loggers to be "skipped")
            logger.removeHandler(self)
        self._loggers.clear()

    def subhandler(self, *args, **kwargs) -> 'MultiStreamFileHandler':
        '''Generate a subordinate "child" process which reports back up to the spawning "parent" process once complete'''
        child = self.__class__(*args, **kwargs) # generalizes inheritance much better
        self.children[child.id] = child # register the child for reference
        child.parent = self # assert self as the child's parent (allows child to tell if any parent handlers exist above it)

        return child
    
    def propogate_msg(self, level : int, msg : str) -> None:
        '''Propogate a logged message up through the parent tree'''
        self.personal_logger.log(level=level, msg=msg) # log message at the determined appropriate level
        if self.parent is not None:                           # if the current process is the child of another...
            self.parent.propogate_msg(level=level, msg=msg)   # ...pass the completion message up the chain to the parent

    def __enter__(self) -> 'MultiStreamFileHandler':
        self._start_time = datetime.now()
        return self
    
    def __exit__(self, exc_type, exc, trace) -> bool:
        if exc_type is not None: # unexpected error
            completion_msg = ''.join(format_exception(exc_type, exc, trace)) # format error message and traceback similar to console printout
            log_level = logging.FATAL
        else: # normal completion of context block
            completion_msg = f'Process "{self.proc_name}" completed in {datetime.now() - self._start_time}\n'
            log_level = logging.INFO

        self.propogate_msg(level=log_level, msg=completion_msg) # log message at the determined appropriate level, passing up to parents if necessary
        # if self.parent is not None:                                               
        #     self.parent.children.pop(self.id) # orphan the current handler once its process is complete
        self.unregister_loggers() # prevents multiple redundant writes within the same Python session

        return True # TOSELF : return Falsy value only if errors are unhandled

class MultiStreamFileHandlerFlexible(MultiStreamFileHandler):
    '''
    MSFHandler which is a bit more flexible (and lazy) when it comes to log file naming and creation
    Can either pass a log file path (as normal) or a directory into which to output the logfile
    If the latter is chosen, will create a logfile based on the specified process name, optionally adding a timestamp if desired
    '''
    def __init__(self, filedir : Optional[Path]=None, filename : Optional[Union[str, Path]]=None, mode : str='a', encoding : Optional[str]=None, delay : bool=False, errors : Optional[str]=None, # FileHandler base args
                 loggers : Optional[Union[str, Logger, list[Logger]]]='ALL', formatter : logging.Formatter=LOG_FORMATTER, proc_name : str='Process', write_timestamp : bool=True, timestamp : Timestamp=Timestamp()) -> None: # new args specific to this class
        if filename is None:
            if filedir is None:
                raise AttributeError('Must specify either a path to log file OR an output directory in which to generate a log file')
            assert(filedir.is_dir()) # double check that a proper directory has in fact been passed
            
            # implicit "else" if no error is raised 
            filestem = proc_name.replace(' ', '_') # use process name to generate filename; change spaces to underscores for file saving
            if write_timestamp:
                filestem += f'_{timestamp.timestamp_now()}' 
            filename = filedir / f'{filestem}.log'

        super().__init__(filename, mode, encoding, delay, errors, loggers, formatter, proc_name)

# ALIASES TO MAKE CLASS NAMES LESS UNWIELDY
MSFHandler        = MultiStreamFileHandler
MSFHandlerFlex    = MultiStreamFileHandlerFlexible


'''Unit tests for tree-related functionality'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Any
import pytest

from anytree.node import Node
from anytree.search import find as find_node
from anytree.iterators import AbstractIter, PreOrderIter, PostOrderIter, LevelOrderIter

from networkx import dfs_preorder_nodes

from polymerist.genutils.trees.treecopy import get_node_attrs, copy_node_unbound, copy_tree, tree_to_networkx


@pytest.fixture
def example_tree() -> Node: 
    '''Produce a simplified tree for performing tests'''
    root = Node('f')
    b = Node('b', parent=root, foo='bb')
    a = Node('a', parent=b,    foo='aa', bar='spam')
    d = Node('d', parent=b,    foo='dd')
    c = Node('c', parent=d,    foo='cc', bar='spam')
    e = Node('e', parent=d,    foo='ee', baz=[1,2,3])
    g = Node('g', parent=root, foo='gg', bar='spam', baz=[1,2,3])
    i = Node('i', parent=g,    foo='ii')
    h = Node('h', parent=i,    foo='hh')
    a_dup = Node('a', foo='a+a', parent=g) # testing how nodes with duplicate names are handled

    return root

@pytest.fixture
def g_node(example_tree) -> Node:
    '''Returns a particularly attribute-rich and neighbor-rich node for testing'''
    return find_node(example_tree, filter_=lambda node : node.name == 'g')

@pytest.fixture
def node_name() -> str:
    '''Allows for configuration of test name values (test should pass with arbitrary names)'''
    return 'test'

@pytest.fixture
def node_attrs() -> dict[str, Any]:
    '''Allows for configuration of test attr values (test should pass with arbitrary attrs)'''
    return {
        'foo'  : 'bar',
        'baz'  : 42,
        'spam' : 'eggs',
    }

def test_get_node_attrs_no_name(node_name : str, node_attrs : dict[str, Any]) -> None:
    '''Test that correct Node attributes are fetched (excluding Node name)'''
    node = Node(name=node_name, **node_attrs)
    attrs = get_node_attrs(node, include_name=False)
    assert attrs == node_attrs

def test_get_node_attrs_with_name(node_name : str, node_attrs : dict[str, Any]) -> None:
    '''Test that correct Node attributes are fetched (including Node name)'''
    node = Node(name=node_name, **node_attrs)
    attrs = get_node_attrs(node, include_name=True)
    assert attrs == {'name' : node_name, **node_attrs}

def test_get_node_attrs_attr_name_filter(node_name : str, node_attrs : dict[str, Any]) -> None:
    '''Test that filter conditions are correctly applied when fetching Node attributes'''
    node = Node(name=node_name, **node_attrs)
    attrs = get_node_attrs(node, attr_filter=lambda attr_name : attr_name != 'foo', include_name=False)
    assert 'foo' not in attrs


def test_copy_node_unbound_values(g_node : Node) -> None:
    '''Test that copy_node_unbound() correctly copies Node attributes'''
    copied_node = copy_node_unbound(g_node)
    assert get_node_attrs(copied_node) == get_node_attrs(g_node)

def test_copy_node_unbound_relatives(g_node : Node) -> None:
    '''Test that copy_node_unbound() correctly removes ancestors and children from node'''
    copied_node = copy_node_unbound(g_node)
    assert copied_node.parent is None and not copied_node.children


@pytest.mark.parametrize('iter_type', [PreOrderIter, PostOrderIter, LevelOrderIter])
def test_copy_tree(example_tree : Node, iter_type : AbstractIter) -> None:
    '''Test that trees structures are exactly copied with no filters'''
    copied_tree = copy_tree(example_tree, stop=None, attr_filter=None)
    assert all(
        get_node_attrs(node_orig) == get_node_attrs(node_copied) 
            for node_orig, node_copied in zip(iter_type(example_tree), iter_type(copied_tree))
    )

@pytest.mark.parametrize('iter_type', [PreOrderIter, PostOrderIter, LevelOrderIter])
def test_copy_tree_mod_attr(example_tree : Node, iter_type : AbstractIter) -> None:
    '''Test that modifying attributes on copied trees does NOT affect original'''
    copied_tree = copy_tree(example_tree, stop=None, attr_filter=None)
    TARG_ATTR : str = 'foo'
    for node in iter_type(copied_tree): # modify attributes
        if hasattr(node, TARG_ATTR):
            setattr(node, TARG_ATTR, getattr(node, TARG_ATTR) + '__')

    assert all( # attributes should NOT be equal, since they should've only been changed on the copy
        getattr(node_orig, TARG_ATTR) != getattr(node_copied, TARG_ATTR)
            for node_orig, node_copied in zip(iter_type(example_tree), iter_type(copied_tree))
                if hasattr(node_orig, TARG_ATTR)
    )

def test_copy_tree_stop(example_tree : Node) -> None:
    '''Test that stop conditions for tree copying are respected and targetted branches are pruned'''
    copied_tree = copy_tree(example_tree, stop=lambda node : node.name == 'a', attr_filter=None)
    assert all(
        copied_node.name != 'a'
            for copied_node in PreOrderIter(copied_tree) # in this case, the iteration order uniquely doesn't matter, only care that all nodes are traversed 
    )


def test_tree_to_networkx(example_tree : Node) -> None:
    '''Test that conversion to networkx.DiGraph faithfully reproduces node order and attributes'''
    nxtree = tree_to_networkx(example_tree)
    assert all(
        nxtree.nodes[i] == get_node_attrs(at_node, include_name=True)
            for i, at_node in zip(dfs_preorder_nodes(nxtree), PreOrderIter(example_tree)) # preorder should match, regardless of implementation
    )


'''Functionality for dynamically importing and inspecting Python modules and packages'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from .pkgiter import module_hierarchy, iter_submodules, module_tree, module_tree_direct
from .pkginspect import is_package, is_module


'''Tools for manipulating files and directories in the file system'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Iterable

from pathlib import Path
from subprocess import Popen

from ..trees.treebase import NodeCorrespondence, compile_tree_factory
from ..trees.treeviz import treestr
from ..decorators.functional import allow_string_paths


# FILE TREES
class PathToNodeCorrespondence(NodeCorrespondence, FROMTYPE=Path):
    '''Concrete implementation of pathlib Paths as nodes in a tree'''
    def name(self, path : Path) -> str:
        return path.name
    
    def has_children(self, path : Path) -> bool:
        return path.is_dir()
    
    def children(self, path) -> Iterable[Path]:
        return path.iterdir()
    
path_tree = file_tree = allow_string_paths(
    compile_tree_factory(
        PathToNodeCorrespondence(),
        obj_attr_name='ppath' # NOTE: can't call this "path", as that clashes with an attribute of Node also called "Path"
    )
)
dir_tree = allow_string_paths(
    compile_tree_factory(
        PathToNodeCorrespondence(),
        class_alias='directory',
        obj_attr_name='dir',
        exclude_mixin=lambda path : path.is_file()
    )
)

# MODIFICATIONS TO DIRECTORIES ON DISC
def startfile(path : Path) -> None:
    '''Replacement for os.startfile() functionality, since none natively exists in Linux'''
    Popen(['xdg-open', path])

def clear_dir(path : Path) -> None:
    '''Recursively clear contents of a directory at the given path (depth-first)'''
    assert(path.is_dir())

    for sub_path in path.iterdir():
        if sub_path.is_dir():
            clear_dir(sub_path)
            sub_path.rmdir() # raises OSError if inside of target subfolder while being deleted
        else:
            sub_path.unlink()



'''SMARTS-based queries for functional groups and other chemical signatures'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

# Creating module-specific logger
import logging
LOGGER = logging.getLogger(__name__)

# locating Path to module for SMARTS data I/O
from pathlib import Path
_MODULE_PATH = Path(__file__).parent

import pandas as pd

from ._daylight_scrape import scrape_SMARTS, FnGroupSMARTSEntry


_fgtab_name : str = 'fn_group_smarts'
_fgtab_path = _MODULE_PATH / f'{_fgtab_name}.csv'

if not _fgtab_path.exists(): # if data table is missing, scrape data back off of Daylight SMARTS sight and save
    LOGGER.warning(F'No functional group SMARTS data from LUT found on system; regenerating from {_daylight_scrape.DAYLIGHT_URL}')
    FN_GROUP_TABLE = _daylight_scrape.scrape_SMARTS()
    FN_GROUP_TABLE.sort_values('category', inplace=True)  # sort by category
    FN_GROUP_TABLE.reset_index(drop=True, inplace=True)   # renumber by new order
    FN_GROUP_TABLE.to_csv(_fgtab_path, index=False) # save for future lookup
else:
    LOGGER.info('Loading functional group SMARTS data from LUT')
    FN_GROUP_TABLE = pd.read_csv(_fgtab_path) # otherwise, read data from table on import

FN_GROUP_ENTRIES = [ # collate into list of easily queryable entries
    FnGroupSMARTSEntry(**fgdict)
        for fgdict in FN_GROUP_TABLE.to_dict(orient='records')
]


'''Tools for labelling and extracting force groups for Forces within an OpenMM System'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Union
from collections import defaultdict
from openmm import Force, NonbondedForce, System


# HANDLING FORCE GROUPS
def force_groups_are_unique(system : System) -> bool:
    '''Check whether the Forces in an OpenMM system share their force group ID with any other forces'''
    seen = set()
    for force in system.getForces():
        force_group = force.getForceGroup()
        if force_group in seen:
            return False
        seen.add(force_group)
    else:
        return True

def uniquify_force_groups(system : System) -> None:
    '''Assigns each Force in an OpenMM System with a unique force group index'''
    for i, force in enumerate(system.getForces()):
        force.setForceGroup(i)
    # TODO : add labelling (depends partially on Interchange's NonbondedForce separation)

def impose_unique_force_groups(ommsys : System) -> None:
    '''Impose unique labels on Forces in an OpenMM System'''
    if not force_groups_are_unique(ommsys):
        uniquify_force_groups(ommsys)

def forces_by_force_group(system : System, denest : bool=False) -> dict[int, Union[Force, list[Force]]]:
    '''Compile the Forces in an'''
    force_dict = defaultdict(list)
    for force in system.getForces():
        force_dict[force.getForceGroup()].append(force)
    force_dict = dict(force_dict) # convert to vanilla dict to avoid non-builtin covariant typing
    
    if denest:
        for group_id, force_list in force_dict.items():
            if len(force_list) == 1: # for single-force lists, extract item and make value directly
                force_dict[group_id] = force_list.pop() 
    
    return force_dict


'''For dynamically determining and cataloging which SMIRNOFF-copatible force fields are installed (and accompanying functionality) are available'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Optional
from pathlib import Path

from ...genutils.importutils.dependencies import modules_installed


# Force field and ToolkitWrapper reference
FFDIR : Optional[Path] = None
if modules_installed('openff.toolkit'):
    from openforcefields import get_forcefield_dirs_paths
    
    FFDIR = Path(get_forcefield_dirs_paths()[0]) # Locate path where OpenFF forcefields are installed

FF_DIR_REGISTRY  : dict[Path, Path] = {}
FF_PATH_REGISTRY : dict[Path, Path] = {}
if modules_installed('openforcefields'):
    from openff.toolkit.typing.engines.smirnoff.forcefield import _get_installed_offxml_dir_paths
    
    for ffdir_str in _get_installed_offxml_dir_paths():
        ffdir = Path(ffdir_str)
        ffdir_name = ffdir.parent.stem

        FF_DIR_REGISTRY[ ffdir_name]  = ffdir
        FF_PATH_REGISTRY[ffdir_name] = [path for path in ffdir.glob('*.offxml')]


'''Custom exceptions specific to RDKit-related functionality''' 

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

# TODO : consider divvying up this module down if few other modules import individual errors
class SubstructMatchFailedError(Exception):
    '''Raised when molecule graph isomorphism match does not form a cover'''
    pass

class BondOrderModificationError(Exception):
    '''Raised when an invalid RDKit bond modification is attempted'''
    pass


'''For obtaining info from and for labelling individual RDKit Atoms'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Callable, Generator
from rdkit.Chem.rdchem import Atom

# NEIGHBOR ATOM INFO
def get_num_bonds(atom : Atom) -> int:
    '''Returns number of explicit bonded connections an atom has (distinct from atom.GetExplicitValence due to bond orders)'''
    return len(atom.GetBonds())

def _get_neighbor_factory_by_condition(condition : Callable[[Atom], bool]) -> Callable[[Atom], Generator[Atom, None, None]]:
    '''Factory function for generating neighbor-search functions over Atoms by a boolean condition'''
    def neighbors_by_condition(atom : Atom) -> Generator[Atom, None, None]:
        '''Generate all neighboring atoms satisfying a condition'''
        for nb_atom in atom.GetNeighbors():
            if condition(nb_atom):
                yield nb_atom

    return neighbors_by_condition

def _has_neighbor_factory_by_condition(condition : Callable[[Atom], bool]) -> Callable[[Atom], bool]:
    '''Factory function for generating neighbor-search functions over Atoms by a boolean condition'''
    def has_neighbors_by_condition(atom : Atom) -> bool:
        '''Identify if any neighbors of an atom satisfy some condition'''
        return any(
            condition(nb_atom)
                for nb_atom in atom.GetNeighbors()
        )

    return has_neighbors_by_condition


'''Automata for reading SMIDGE strings into their graph representations'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER = logging.getLogger(__name__)

from typing import Optional, Sequence, Union
from dataclasses import dataclass, field
from abc import ABC, abstractmethod

import networkx as nx

from .smidgebonds import MonomerGraphBondInfo
from ..monographs import MonomerGraph


@dataclass
class SMIDGEWriterRegister:
    '''Information accessed by machine states during the SMIDGE writing process'''
    monograph    : MonomerGraph         = field(default_factory=MonomerGraph) # this is the only initializable field
    bond_info    : MonomerGraphBondInfo = field(default_factory=MonomerGraphBondInfo)
    tokens       : list[int]            = field(default_factory=list)

    curr_idx     : int                  = field(default_factory=int)
    visited      : set                  = field(default_factory=set)
    node_stack   : list[int]            = field(default_factory=list)
    successors   : dict[int, list[int]] = field(default_factory=dict)
    predecessors : dict[int, int]       = field(default_factory=dict)

    finished     : bool                 = field(default_factory=bool)

    @property
    def prev_idx(self) -> Optional[int]:
        '''The index of the precursor to the currently active node'''
        return self.predecessors.get(self.curr_idx)
    
    @property
    def node_successors(self) -> list[int]:
        '''The nodes which follow the current node in the DFS traversal (could be empty)'''
        return self.successors.get(self.curr_idx)


# STATE BASE
class SMIDGEWriterState(ABC):
    '''Abstract base for automaton states used in writing MID graphs to SMIDGE strings'''
    # NOTE: state_action() is deliberately NOT an abstract method, as doing nothing while in a state is perfectly valid
    def state_action(self, register : SMIDGEWriterRegister) -> 'SMIDGEWriterState':
        pass

    @abstractmethod
    def transition(self, register : SMIDGEWriterRegister) -> 'SMIDGEWriterState':
        '''Define which states should follow the current one based on input'''
        pass

# CONCRETE STATES
class CheckIsVisitedNode(SMIDGEWriterState):
    '''Read a node index off the stack and check if it has already been visited'''
    def state_action(self, register: SMIDGEWriterRegister) -> SMIDGEWriterState:
        register.curr_idx = register.node_stack.pop()

    def transition(self, register: SMIDGEWriterRegister) -> SMIDGEWriterState:
        if register.curr_idx in register.visited:
            return DetermineNodeSuccessors()
        else:
            return VisitNode()

class VisitNode(SMIDGEWriterState):
    '''Visit a node, check if it has predecessors, and mark it as visited'''
    def state_action(self, register: SMIDGEWriterRegister) -> SMIDGEWriterState:
        if register.prev_idx is not None:
            register.bond_info = MonomerGraphBondInfo(
                incoming_flavor=register.monograph.nodes[register.prev_idx].get(MonomerGraph.FLAVOR_DICT_ATTR, {}).get(register.curr_idx),
                bondtype=register.monograph.edges[register.prev_idx, register.curr_idx].get(MonomerGraph.BONDTYPE_ATTR),
                outgoing_flavor=register.monograph.nodes[register.curr_idx].get(MonomerGraph.FLAVOR_DICT_ATTR, {}).get(register.prev_idx), # NOTE : the same as incoming_flavor, but with the order of nodes reversed
            )
        else:
            register.bond_info = MonomerGraphBondInfo()
        register.visited.add(register.curr_idx) # mark the current node as visited

    def transition(self, register: SMIDGEWriterRegister) -> SMIDGEWriterState:
        if register.prev_idx is not None:
            return WriteEdge()
        else:
            return WriteNode()

class WriteNode(SMIDGEWriterState):
    '''Add an entry for the current node to the output tokens'''
    def state_action(self, register: SMIDGEWriterRegister) -> SMIDGEWriterState:
        monomer_token = f'[{register.monograph.nodes[register.curr_idx][register.monograph.MONOMER_NAME_ATTR]}]'
        register.tokens.append(monomer_token)
        
        if not register.successors:
            register.finished = True # will implicitly cause this to be the final machine state when parsin
    
    def transition(self, register: SMIDGEWriterRegister) -> SMIDGEWriterState:
        return DetermineNodeSuccessors()

class WriteEdge(SMIDGEWriterState):
    '''Add an entry for the current edge to the output tokens'''
    def state_action(self, register: SMIDGEWriterRegister) -> SMIDGEWriterState:
        register.tokens.append(f'<{register.bond_info!s}>')

    def transition(self, register: SMIDGEWriterRegister) -> SMIDGEWriterState:
        return WriteNode()

class DetermineNodeSuccessors(SMIDGEWriterState):
    '''Determine branching and next-node behavior'''
    def transition(self, register: SMIDGEWriterRegister) -> SMIDGEWriterState:
        if register.node_successors:
            if len(register.node_successors) > 1:
                return BranchStart()
            else:
                return PushNodeSuccessor()
        else:
            if register.successors:
                return BranchEnd() # always return a branch end condition
            else:
                return UpdateNodeSuccessors() # EXCEPT when the current node is also the last node

class PushNodeSuccessor(SMIDGEWriterState):
    '''Add a nodes DFS successor to the traversal stack'''
    def state_action(self, register: SMIDGEWriterRegister) -> SMIDGEWriterState:
        register.node_stack.append(register.node_successors.pop(0))
    
    def transition(self, register: SMIDGEWriterRegister) -> SMIDGEWriterState:
        return UpdateNodeSuccessors()

class UpdateNodeSuccessors(SMIDGEWriterState):
    '''Remove any node with no successors from the DFS traversal map'''
    def state_action(self, register: SMIDGEWriterRegister) -> SMIDGEWriterState:
        if not register.node_successors:
            register.successors.pop(register.curr_idx, None) # remove the current node from the map of possible successors (None default silences KeyError if only implicitly present)

    def transition(self, register: SMIDGEWriterRegister) -> SMIDGEWriterState:
        return CheckIsVisitedNode()

class BranchStart(SMIDGEWriterState):
    '''Mark position of branch point in stack for backtrack'''
    def state_action(self, register: SMIDGEWriterRegister) -> SMIDGEWriterState:
        register.node_stack.append(register.curr_idx) # push the current node onto the stack again to mark for revisiting later
        register.tokens.append('(')

    def transition(self, register: SMIDGEWriterRegister) -> SMIDGEWriterState:
        return PushNodeSuccessor()

class BranchEnd(SMIDGEWriterState):
    '''Mark the end of a branch'''
    def state_action(self, register: SMIDGEWriterRegister) -> SMIDGEWriterState:
        register.tokens.append(')')

    def transition(self, register: SMIDGEWriterRegister) -> SMIDGEWriterState:
        return UpdateNodeSuccessors()


# WRITER
@dataclass
class SMIDGEWriter:
    '''Register machine for translating SMIDGE strings to and from monomer graphs'''
    state : SMIDGEWriterState = field(default_factory=CheckIsVisitedNode, init=False) # initial state is always the new chain state

    def write_smidge_connected_component(self, monograph : MonomerGraph, start_node_idx : Optional[int]=None) -> MonomerGraph:
        '''Parse a single connected component of a monomer graph into a SMIDGE string'''
        self.state = CheckIsVisitedNode()
        register = SMIDGEWriterRegister(
            monograph=monograph,
            curr_idx=start_node_idx,
            successors=nx.dfs_successors(monograph, source=start_node_idx),
            predecessors=nx.dfs_predecessors(monograph, source=start_node_idx),
            node_stack=[start_node_idx],
            finished=False
        )

        while not register.finished:
            LOGGER.debug(f'Current {register.__class__.__name__}: {register!r}')
            LOGGER.debug(f'Current {self.__class__.__name__} state: {self.state.__class__.__name__}')
            self.state.state_action(register)
            self.state = self.state.transition(register)
        return ''.join(register.tokens)
    
    def write_smidge(self, monograph : MonomerGraph, start_node_idxs :  Optional[Union[int, Sequence[int]]]=None) -> str:
        '''Parse all connected components of a monograph into a complete SMIDGE string'''
        chain_starts = monograph._validate_start_node_idxs(start_node_idxs)
        return '.'.join(
            self.write_smidge_connected_component(chain, start_node_idx=chain_starts[i])
                for i, chain in enumerate(monograph.chains)
        )


'''Utilites for representing pure permutations, cycles, and permutation groups'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Generator, Iterable, Optional, Sequence, TypeVar
from dataclasses import dataclass, field
T = TypeVar('T')

import numpy as np
from math import lcm
from operator import mul

from functools import reduce
from itertools import chain, permutations
from collections import Counter, defaultdict


class Cycle(tuple):
    '''For representing a cyclic collection of objects'''
    def __new__(cls, *values : Iterable[T]) -> None: # override new to allow for star unpacking
        return super(Cycle, cls).__new__(cls, values)

    def __repr__(self) -> str:
        '''Represent as a '''
        return f'{self.__class__.__name__}{super().__repr__()}'

    def __getitem__(self, __key : int) -> T:
        '''Access an item, with wrap-around indices'''
        return super().__getitem__(__key % len(self))

    def __reversed__(self) -> 'Cycle':
        return self.__class__(*super().__reversed__())

    def starting_from_index(self, index : int) -> 'Cycle':
        '''Return a Cycle which is in the same order as self but begins at the given index'''
        return Cycle(*(self[i] for i in range(index, index + len(self))))

    def __add__(self, other : 'Cycle') -> 'Cycle':
        '''Extend a Cycle by another Cycle'''
        if not isinstance(other, Cycle):
            raise TypeError
        return self.__class__(*super().__add__(other))

    def __call__(self, other) -> 'Cycle': # TODO : implement composition via R-to-L adjacency
        ...

    def copy(self) -> 'Cycle':
        return self.__class__(*self)

    @property
    def mapping(self) -> dict[T, T]:
        '''A mapping of each item in the cycle to the following item'''
        return {
            self[i - 1] : elem # NOTE: implemented this way specifically to make use of "-1" last index and avoid and IndexError for the wrap-around
                for i, elem in enumerate(self)
        }

    # HELPER METHODS
    @staticmethod
    def max_elem_in_cycles(cycles : Iterable['Cycle']) -> int:
        '''Returns the largest element across a collection of Cycles'''
        return max(chain(*cycles))

    @staticmethod
    def cycles_are_disjoint(cycles : Iterable['Cycle']) -> bool:
        '''Check if a cycle decomposition is disjoint'''
        return set.intersection(*map(set, cycles)) == set() # check that the intersection of all cycles is the empty set

    @staticmethod
    def cycles_produce_partition(cycles : Iterable['Cycle']) -> bool:
        '''Check if a cycle decomposition forms a partition of some set of integers'''
        degree = Cycle.max_elem_in_cycles(cycles) + 1 # find total maximum to use as order
        all_elems = set.union(*map(set, cycles)) # check that the intersection of all cycles is the empty set

        for i in range(degree):
            if i not in all_elems:
                return False
        else:
            return True

    @staticmethod
    def cycle_type(cycles : Iterable['Cycle']) -> dict[int, int]:
        '''Returns a dict of cycle lengths and the number of cycle in the permutation with that length'''
        cycle_len_counts = Counter()
        for cycle in cycles:
            cycle_len_counts[len(cycle)] += 1
        longest_cycle_len = max(cycle_len_counts.keys())

        return {
            cycle_len : cycle_len_counts[cycle_len]
                for cycle_len in range(1, longest_cycle_len + 1) # include zeros for intermediate sizes, 1-index to make use of counts
        }

    @staticmethod
    def cycle_index_sym(cycle_type : dict[int, int], var_sym : str='x', mul_sym : str='*', exp_sym : str='^') -> str: # TODO : add support for excluding x_n**0 (i.e. null) cycles
        '''Represent a single cycle type in symbolic variable form
        Can optionally supply custom symbols for variable (single-char only), multiplication sign, and exponentiation sign'''
        if len(var_sym) > 1:
            raise ValueError('must provide sinle-character symbolic variable for cycle index')
        return mul_sym.join(f'{var_sym}_{cycle_len}{exp_sym}{count}' for cycle_len, count in cycle_type.items())


@dataclass # this is purely to give nice bool, repr, and eq boilerplates
class Permutation:
    '''For representing a permutation'''
    elems : Sequence[int] = field(default_factory=list)

    def __init__(self, *elems : Sequence[int]) -> None: # need to implement __init__ in order to allow star unpacking
        self.elems = elems
        if self.element_set != set(self.natural_order):
            raise ValueError

    # CARDINALITY
    @property
    def degree(self) -> int:
        return len(self.elems)

    def __len__(self) -> int:
        return self.degree
    
    @property
    def element_set(self) -> set[int]:
        return set(self.elems)
    
    @staticmethod
    def _natural_order(N : int) -> list[int]:
        '''The first N natural numbers {0, ... ,N-1} in ascending order'''
        return [i for i in range(N)]

    @classmethod
    def from_degree(cls, degree : int, random : bool=False) -> 'Permutation':
        '''Construct a permutation of the given degree. Is just the identity by default, but can be made random with the "random" flag'''
        elems = cls._natural_order(degree)
        if random:
            np.random.shuffle(elems)
        return cls(*elems)
    
    @classmethod
    def identity(cls, degree : int) -> 'Permutation':
        '''Construct the Identity permutation of a given degree'''
        return cls.from_degree(degree, random=False)

    # ELEMENTARY PERMUTATIONS
    @property
    def natural_order(self) -> list[int]:
        return self._natural_order(self.degree)
    
    @property
    def as_identity(self) -> 'Permutation':
        '''The identity permutation with the same degree as the current permutation'''
        return self.__class__.identity(self.degree)

    @property
    def inverse(self) -> 'Permutation':
        '''The permutation which, when composed with this permutation from either the left or the right, yields the identity permutation'''
        return self.__class__(*np.argsort(self.elems))
    
    @property
    def reverse(self) -> 'Permutation':
        '''The reversed-order of a permutation'''
        return self.__class__(*reversed(self.elems))
    
    def copy(self) -> 'Permutation':
        '''Create a copy of the current Permutation'''
        return self.__class__(*self.elems)
    
    # COMPOSITIONS AND MAPS
    def image(self, coll : Iterable[T]) -> Iterable[T]:
        '''The image of an ordered collection under the defined permutation'''
        # if len(coll) != self.degree:
        #     raise ValueError
        return [coll[i] for i in self.elems]
    
    def __call__(self, elem : int) -> int:
        '''Return the image of a single element under the permutation'''
        if not isinstance(elem, int):
            raise TypeError
        
        if (elem < 0) or (elem > (self.degree - 1)):
            raise ValueError(f'Integer {elem} has no image under permutation of degree {self.degree}')
        
        return self.elems[elem]
    
    def __getitem__(self, elem : int) -> int:
        return self.elems[elem]

    def __mul__(self, other : 'Permutation') -> 'Permutation':
        if not isinstance(other, Permutation):
            raise TypeError(f'Cannot compose {self.__class__.__name__} with {other.__class__.__name__}')
        return self.__class__(*self.image(other.elems))
    compose = __mul__

    def __pow__(self, exp : int) -> 'Permutation':
        if exp == 0:
            return self.__class__(*self.elems) # identity composition, return copy of self to avoid mutation
        elif exp < 0:
            return self.inverse.__pow__(abs(exp))
        else:
            return reduce(mul, [self for _ in range(exp)]) # TODO : see if there's a more efficient way to do this
        
    @property
    def order(self) -> int:
        '''Smallest power of a permutation which generates the identity permutation'''
        return lcm(*(len(cycle) for cycle in self.to_cycles()))

    # MATRICES
    @staticmethod
    def _is_valid_permutation_matrix(matrix : np.ndarray) -> bool:
        return (                                  # A valid permutation matrix must be:
            (matrix.ndim == 2)                        # 1) 2-dimensional
            and (matrix.shape[0] == matrix.shape[1])  # 2) square
            and (np.in1d(matrix, [0 ,1])).all()       # 3) contain ONLY ones and zeroes
            and (matrix.sum(axis=0) == 1).all()       # 4) have exactly a single 1 in each column
            and (matrix.sum(axis=1) == 1).all()       # 4) have exactly a single 1 in each row
        )

    @classmethod
    def from_matrix(cls, matrix : np.ndarray) -> 'Permutation':
        '''Create a permutation from a permutation matrix'''
        if not cls._is_valid_permutation_matrix(matrix):
            raise ValueError('Provided matrix is not a valid permutation matrix')
        
        _, elems = np.nonzero(matrix.T)
        return cls(*elems)
        # _, elems = np.nonzero(matrix) # alternate, equally-valid implementation
        # return cls(*elems).inverse
    
    def to_matrix(self) -> np.ndarray:
        '''Obtain permutation matrix representation'''
        return np.eye(self.degree, dtype=int)[:, self.elems]
    
    @property
    def matrix(self) -> np.ndarray:
        return self.to_matrix()

    # WORD FORMS
    @classmethod
    def from_word(cls, word : str, delimiter : str='') -> 'Permutation':
        '''Create a permutation from a string with a total ordering'''
        return cls(*np.argsort(word.split(delimiter))).inverse # need to invert to produce expected result, as argsort gives the permutation which restores the identity

    def to_word(self) -> str:
        return ' '.join(str(i) for i in self.elems) # TODO : find best way to deal with multi-digit numbers

    # CYCLES
    def _cycle_decomposition(self) -> list[Cycle]:
        '''Return the disjoint cycles of a Permutation'''
        visited : list[bool] = [False] * self.degree
        cycles = []
        for elem in self.elems:
            cycle_elems = []
            while not visited[elem]:
                cycle_elems.append(elem)
                visited[elem] = True
                elem = self.elems[elem]
            if cycle_elems:
                cycles.append(Cycle(*cycle_elems))

        return cycles
    
    @classmethod
    def from_cycle(cls, cycle : Cycle, degree : Optional[int]=None) -> None: 
        '''Create a permutation from an ordering of elements in a single cycle. Missing elements are inferred to be fixed points'''
        max_elem = max(cycle) # determine value once to avoid multiple calls
        if degree is None:
            degree = max_elem
        elif (max_elem > degree - 1): # check if the degree is provided BUT is incompatible with the natural number set provided
            raise ValueError(f'Cannot construct a permutation of degree {degree} for a set containing at least {max_elem + 1} elements')
        return cls(*(cycle.mapping.get(i, i) for i in range(degree + 1))) # insert mapped element if provided, or else one with the same index (i.e. a fixed point)

    @classmethod
    def from_cycles(cls, cycles : list[Cycle]) -> 'Permutation':
        '''Stitch together Permutation from disjoint cycle representation'''
        degree = Cycle.max_elem_in_cycles(cycles) + 1
        return reduce(
            mul,
            (cls.from_cycle(cycle, degree=degree) for cycle in cycles[::-1]), # apply cycles right-to-left
            cls.identity(degree) # initialize with identity permutation
        ) 
    
    def as_cycle(self) -> Cycle:
        '''Reinterpret permutation a defining a cyclic order'''
        return Cycle(*self.elems)

    def to_cycles(self, canonicalize : bool=True) -> list[Cycle]:
        '''Produce cycle decompositon of permutation. NOT TO BE CONFUSED WITH Permutation.as_cycles()!
        By default, returns in canonical order (i.e. each cycle is presented with the least element first)'''
        if not canonicalize:
            return self._cycle_decomposition()
        
        # implicit else for canonicalization of cycles
        return sorted( # in canonical cycle notation, cycles are:
            (cycle.starting_from_index(np.argmax(cycle)) for cycle in self._cycle_decomposition()), # 1) listed with the largest element first          
            key=lambda x : x[0]                                                            # 2) sorted in ascending order by the first element
        ) 

    @property
    def cycles(self) -> list[Cycle]:
        '''Cycle decompositon of permutation in the order that elements appear (i.e. NOT in canonical order)'''
        return self.to_cycles(canonicalize=False)

    @property
    def cycles_canonical(self) -> list[Cycle]:
        '''Cycle decompositon of permutation in canonical order (largest element) first in each cycle and)'''
        '''Get cycle decompositon of permutation is canonical order'''
        return self.to_cycles(canonicalize=True)
    standard_notation = standard_form = canonical_cycles = cycles_canonical # lots of aliases to avoid getting bogged down in notation

    @property
    def cycle_type(self) -> dict[int, int]:
        '''Returns a dict of cycle lengths and the number of cycle in the permutation with that length'''
        return Cycle.cycle_type(self.to_cycles())

    # INVERSIONS AND LEHMER CODES - TODO: implement enumeration with factoradics
    @classmethod
    def from_lehmer(cls, lehmer_code : Sequence[int]) -> 'Permutation':
        '''Stitch together permutation from an inversion vector'''
        elems = [elem for elem in reversed(lehmer_code)] # create reversed copy to simplify iteration indexing avoid modifying original lehmer code
        for i, elem in enumerate(elems):
            for j, right_elem in enumerate(elems[:i]):
                if right_elem >= elem:
                    elems[j] += 1
        return cls(*reversed(elems)) # reverse final result to recover permutation order

    def to_lehmer_code(self) -> list[int]:
        '''Construct the left lehmer code for a permutation
        Each position in the code gives the number of inversions to the right of the permutation element at the corresponding position'''
        vector = [elem for elem in self.elems] # initialize witha  copy of the positions
        for i, elem in enumerate(vector):
            for j, right_elem in enumerate(vector[i:], start=i): # start indices at i to preserve correct absolute position in list
                if right_elem > elem:
                    vector[j] -= 1
        return vector
    
    @property
    def lehmer_code(self) -> list[int]:
        return self.to_lehmer_code()
    inversion_vector = lehmer = lehmer_code # aliases for convenience

    @property
    def num_inversions(self) -> int:
        '''Get total number of "out-of-order" elements in the permutation'''
        return sum(self.lehmer)

    @property
    def sign(self) -> int:
        '''The parity of the number of inversion of a permutation'''
        return -1 if (self.num_inversions % 2) else 1
    parity = sign

    @property
    def is_even(self) -> bool:
        return (self.sign == 1)

    @property
    def is_odd(self) -> bool:
        return (self.sign == -1)

    # ASCENTS AND DESCENTS
    @property
    def ascents(self) -> list[int]:
        '''The positions of elements which are followed by a greater element'''
        return [
            i for i, elem in enumerate(self.elems[:-1])
                if elem < self.elems[i + 1]
        ]
        
    @property
    def num_ascents(self) -> int:
        '''The number of elements which are followed by a greater element'''
        return len(self.ascents)

    @property
    def descents(self) -> list[int]:
        '''The positions of elements which are followed by a lesser element'''
        return [
            i for i, elem in enumerate(self.elems[:-1])
                if elem > self.elems[i + 1]
        ]
        
    @property
    def num_descents(self) -> int:
        '''The number of elements which are followed by a lesser element'''
        return len(self.descents)

    @property
    def support(self) -> list[int]:
        '''The positions of elemnts which do not map to themselves'''
        return [i for i, elem in enumerate(self.elems) if elem != i]
    
    @property
    def support_size(self) -> int:
        '''The number of elements in the permutation which do not map to themselves'''
        return len(self.support)
    
    # PERMUTATION GROUPS
    @staticmethod
    def cycle_index(perm_group : Iterable['Permutation'], variable : str='x', add_sym : str=' + ', mul_sym : str='*', exp_sym : str='^') -> str:
        '''Construct a string of the cycle index of a permutation group'''
        group_size : int = 0 # this allows passing of Iterable permutation groups and avoid needing to know group size a priori
        cycle_type_counts = defaultdict(int)
        for perm in perm_group:
            cycle_index_str = Cycle.cycle_index_sym(perm.cycle_type, variable, mul_sym, exp_sym)
            cycle_type_counts[cycle_index_str] += 1
            group_size += 1

        return f'1/{group_size}*({add_sym.join(f"{count}{mul_sym}{cycle_index_str}" for cycle_index_str, count in cycle_type_counts.items())})'

    @classmethod
    def symmetric_group(cls, order : int) -> Generator['Permutation', None, None]:
        '''Generate all permutations of a given order'''
        for elems in permutations(range(order), order):
            yield cls(*elems)

    @classmethod
    def alternating_group(cls, order : int) -> Generator['Permutation', None, None]:
        '''Generate all even permutations of a given order'''
        for perm in cls.symmetric_group(order):
            if perm.is_even:
                yield perm

    @classmethod
    def cyclic_group(cls, order : int, _base_perm : Optional['Permutation']=None) -> Generator['Permutation', None, None]:
        '''Generate all cyclic shifts of a given permutation of a given order'''
        if _base_perm is None:
            generator = cls.identity(order) # the identity permutation is the generator of the cyclic group
        elif _base_perm.degree != order:
            raise ValueError(f'Cannot use the {cls.__name__} {_base_perm} of degree {_base_perm.degree} as generator of the Cyclic Group of order {order}')
        else:
            generator = _base_perm

        cycle = generator.as_cycle()
        for i in range(order):
            yield cls(*cycle.starting_from_index(i))

    @classmethod
    def dihedral_group(cls, order : int) -> Generator['Permutation', None, None]:
        '''Generate all permutations of vertex-numbered regular polygon'''
        yield from cls.cyclic_group(order)
        yield from cls.cyclic_group(order, _base_perm=cls.identity(order).reverse) # reflect, then cycle element again


'''Representations and calculation methods for crystallographic unit cells and lattice parameters'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Callable, ClassVar, Optional, Type
from numbers import Number
from ...genutils.typetools.numpytypes import Shape, D
from dataclasses import dataclass, field

import numpy as np
from .integral import nearest_int_coord_along_normal, CubicIntegerLattice
from .coordinates import BoundingBox, Coordinates


def identify_bravais_points_within_bbox(lattice_vectors : np.ndarray[Shape[D, D], Number], bbox : BoundingBox) -> tuple[Coordinates, CubicIntegerLattice]:
    '''Locate all lattice points generated by a set of Bravais lattice vectors in D-dimensions which fall within a given bounding box
    Returns the coordinate vector of the internal lattice points and a CubicIntegerLattice containing the corresponding lattice vector multiplicities'''
    # 1) transform the bounding box into the inverse domain, where Bravais lattice points "look like" integer points
    inv_bbox_coords = bbox.linear_transformation(np.linalg.inv(lattice_vectors.T), as_coords=True)
    directors = (inv_bbox_coords.points - inv_bbox_coords.centroid) # directional vectors at each vertex which are away from transformed box center
    directors /= np.linalg.norm(directors, axis=1)[:, None] # normalize direction vectors

    # 2) find the nearest integer-valued points to transformed bounding box points in directions "away" from the bounding box
    integral_bounding_points = np.vstack([ 
        nearest_int_coord_along_normal(point, director) # ...(ensures no bounded points end up outside of final box)
            for point, director in zip(inv_bbox_coords.points, directors)
    ])

    # 3) produce collection of integer point which contains AT LEAST the integral points bounded by the transformed nearest-integrer bounding points
    trial_int_bbox = BoundingBox(integral_bounding_points)  # TODO: box is pretty inefficient for highly-sheared unit vectors: need simple method to test for integral points within a bounding volume (or simplex) 
    trial_int_latt = CubicIntegerLattice(trial_int_bbox.dimensions + 1) # need K+1 unit-spaced points to space an interval K units long (e.g. [0-2] -> (0, 1, 2))
    trial_int_latt.translate(trial_int_bbox.minimum) # offset lattice from default at origin

    # 4) transforms the trial points back to the normal domain and perform a final enclosure check
    trial_brav_latt = trial_int_latt.linear_transformation(lattice_vectors.T, as_coords=True)
    is_inside = bbox.surrounds(trial_brav_latt).all(axis=1) # mask of which points have ALL coordinates inside the bounding box
    inside_brav_latt      = Coordinates(trial_brav_latt.points[is_inside]) # create new coordinates of just the interior Bravais lattice points
    trial_int_latt.points = trial_int_latt.points[   is_inside]  # screen out outer integral points in-place (due to how CubicIntegerLattice __init__ is set)
    
    return inside_brav_latt, trial_int_latt

# LATTICE VECTORS AND PARAMETERS
@dataclass
class LatticeParameters: # TODO : add support for OpenMM units, add compat w/ lammpstools.lammpseval
    '''For storing the lengths of and the angles between the 3 lattice vectors of a crystallographic unit cell'''
    a : float
    b : float
    c : float

    alpha : float = field(default=np.pi / 2) # make cell orthorhombic by default
    beta  : float = field(default=np.pi / 2) # make cell orthorhombic by default
    gamma : float = field(default=np.pi / 2) # make cell orthorhombic by default

    _LENGTH_ATTRS : ClassVar[tuple[str]] = ('a', 'b', 'c')
    _ANGLE_ATTRS  : ClassVar[tuple[str]] = ('alpha', 'beta', 'gamma')

    # VALIDATION
    def __post_init__(self) -> None:
        '''Validating initialized values'''
        lens_positive, bad_len_name = self.lengths_are_positive
        if not lens_positive:
            raise ValueError(f'{self.__class__.__name__} unit cell lengths must be positive (passed invalid value {bad_len_name}={getattr(self, bad_len_name)})')
        self.reduce_angles()

    @property
    def lengths_are_positive(self) -> tuple[bool, Optional[str]]:
        '''Check whether all axial lengths are well-defined (i.e. positive)'''
        for len_attr in self._LENGTH_ATTRS:
            if (getattr(self, len_attr) <= 0):
                return False, len_attr
        else:
            return True, None
        
    def reduce_angles(self) -> None:
        '''Ensure all angles are within the principal [0, 2*pi) interval'''
        for angle, angle_attr in zip(self.angles, self._ANGLE_ATTRS):
            setattr(self, angle_attr, angle % (2*np.pi)) # ensure angles are within [0, 2*pi)

    # PROPERTY VIEWS
    @property
    def axial_lengths(self) -> np.ndarray[Shape[3], float]:
        '''View of just the axial lengths'''
        return np.array([getattr(self, len_attr) for len_attr in self._LENGTH_ATTRS])
    lengths = axial_lengths

    @property
    def angles(self) -> np.ndarray[Shape[3], float]:
        '''Property alias of angles() method for convenience'''
        return np.array([getattr(self, len_attr) for len_attr in self._ANGLE_ATTRS])
    
    def axial_angles(self, in_degrees : bool=False) -> np.ndarray[Shape[3], float]:
        '''View of just the angles'''
        if in_degrees:
            return np.rad2deg(self.angles)
        return self.angles
    
    @property
    def volume(self) -> float:
        '''The volume of the unit cell (in arbitrary units)'''
        return np.linalg.det(self.lattice_vectors)

    # LATTICE VECTOR METHODS
    @classmethod
    def from_lattice_vectors(cls, vectors : np.ndarray[Shape[3,3], float]) -> 'LatticeParameters':
        '''Obtain axial lengths and inter-vector angles from a matrix of lattice vectors'''
        axial_lengths = np.linalg.norm(vectors, axis=1)
        normed_vectors = vectors / axial_lengths[:, None] # normalize along rows

        BCA = np.roll(normed_vectors, -1, axis=0) # lattice vectors cycled backward
        CAB = np.roll(normed_vectors,  1, axis=0) # lattice vectors cycled forward
        cycled_dots = np.sum(BCA * CAB, axis=1) # row-wise dot product between cycled rows, giving normalized [B.C, C.A, and A.B] as result
        angles = np.arccos(cycled_dots) # invert cosine from dot product expression to recover angles

        return cls(*axial_lengths, *angles)
    
    @staticmethod
    def _lattice_vector_classmethod_factory(lattice_vectors : np.ndarray[Shape[3,3]], lattice_name : str, prefix : str='create') -> Callable[[Type['LatticeParameters'], float], 'LatticeParameters']:
        '''Factory function for generating LatticeParameter initializers based on pre-defined lattice vectors'''
        def generic_lattice_generator(cls, dim : float=1.0) -> LatticeParameters:
            return cls.from_lattice_vectors(dim * lattice_vectors)
        
        # modify signature of generic to mask its dynamic creation
        fn_name = f'{prefix}{"_" if prefix else ""}{lattice_name.lower()}'
        generic_lattice_generator.__name__ = fn_name
        generic_lattice_generator.__doc__  = f'Generate lattice parameters for a {lattice_name.lower()} box (with optional uniform scaling)'
        generic_lattice_generator.__qualname__ = f'{LatticeParameters.__qualname__}.{fn_name}'

        return generic_lattice_generator

    def to_lattice_vectors(self) -> np.ndarray[Shape[3,3], float]:
        '''
        The restricted lattice vectors corresponding to the lattice parameters,
        where vector A lies along the x-axis and vector B is in the xy-plane
        
        Vectors are returned as a 3x3 matrix, where each row represents a lattice vector
        '''
        # formulas adapted from https://www.aflowlib.org/prototype-encyclopedia/triclinic_lattice.html 
        ax = self.a * 1.0                # vector A lies purely along the x-axis
        bx = self.b * np.cos(self.gamma) #    projection onto the x-axis unit vector (i.e. the normalized A vector) via the dot product
        by = self.b * np.sin(self.gamma) # perpendicular onto the x-axis unit vector (i.e. the normalized A vector) via the cross product
        cx = self.c * np.cos(self.beta)  #    projection onto the x-axis unit vector (i.e. the normalized A vector) via the dot product
        cy = self.c * (np.cos(self.alpha) - np.cos(self.beta)*np.cos(self.gamma)) / np.sin(self.gamma) # projection of C onto the y-axis unit vector (via the vector triple product C*(AxB)xA)
        cz = np.sqrt(self.c**2 - cx**2 - cy**2) # remaining z-axis contribution to C must all be along z-axis (algebraically simpler this way)

        return np.array([
            [ax , 0.0, 0.0],
            [bx , by , 0.0],
            [cx , cy , cz ],
        ])

    @property
    def lattice_vectors(self) ->np.ndarray[Shape[3,3], float]:
        '''Property alias of to_lattice_vectors() method for convenience'''
        return self.to_lattice_vectors()
    
## REFERENCE TABLE OF COMMON LATTICE VECTORS WITH UNIT-LENGTH AXIAL VECTORS FOR REFERENCE (WITH DYNAMIC REGISTRATION)
COMMON_UNIT_LATTICE_VECTORS : dict[str, np.ndarray[Shape[3, 3], float]] = {
    'CUBIC' : np.asarray([
        [1.0, 0.0, 0.0],
        [0.0, 1.0, 0.0],
        [0.0, 0.0, 1.0],
    ]),
    'HEXAGONAL' : np.array([
        [ 1.0, 0.0, 0.0],
        [-1/2, np.sqrt(3.0)/2, 0.0],
        [ 0.0, 0.0, 1.0],
    ]),
    'RHOMBOHEDRAL' : np.array([ # equivalent to rhombic dodecahedron xy-hexagon in GROMACS docs
        [1.0, 0.0, 0.0],
        [1/2, np.sqrt(3.0)/2.0, 0.0],
        [1/2, np.sqrt(3.0)/6.0, np.sqrt(6.0)/3.0],
    ]),
    'RHOMBIC_DODECAHEDRON_XY_SQR' : np.array([
        [1.0, 0.0, 0.0],
        [0.0, 1.0, 0.0],
        [1/2, 1/2, np.sqrt(2.0)/2.0],
    ]),
    'TRUNCATED_OCTAHEDRON' : np.array([
        [ 1.0, 0.0, 0.0],
        [ 1/3, 2.0 * np.sqrt(2.0)/3.0, 0.0],
        [-1/3, np.sqrt(2.0)/3.0, np.sqrt(6.0)/3.0],
    ]),
} # adapted from https://manual.gromacs.org/current/reference-manual/algorithms/periodic-boundary-conditions.html

for lattice_name, lattice_vectors in COMMON_UNIT_LATTICE_VECTORS.items():
    globals()[lattice_name] = lattice_vectors # dynamically register common lattice vectors at module level
    
    lattice_generator = LatticeParameters._lattice_vector_classmethod_factory(lattice_vectors, lattice_name=lattice_name)
    setattr(LatticeParameters, lattice_generator.__name__, classmethod(lattice_generator))



'''Specialized cases of general positional numbering systems which are more common in usage'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Callable, Generator, Iterable
from .positional import PositionalNumberingSystem


def hypergeometric_ratios(funct : Callable[[int], int]) -> Generator[int, None, None]:
    '''Generates ratios between successive natural number terms according to a provided function'''
    i = 0
    while True:
        yield funct(i)
        i += 1


class FixedRadixNumberSystem(PositionalNumberingSystem):
    '''Positional numbering system with a single fixed radix'''
    def __init__(self, radix : int=10) -> None:
        self._radix = radix

    @property
    def radix(self) -> int:
        return self._radix
    base = radix

    @property
    def radices(self) -> Iterable[int]: # override radices getter with generator
        return hypergeometric_ratios(lambda i : self.radix)

class FactorialNumberSystem(PositionalNumberingSystem):
    '''For representing factoradic numbers, useful in enmerating permutations via Lehmer codes'''
    def __init__(self) -> None:
        pass

    @property
    def radices(self) -> Iterable[int]:
        return hypergeometric_ratios(lambda i : i + 1)



'''Classes for partial charge assignment of OpenFF Molecules'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER = logging.getLogger(__name__)

from typing import Union
from abc import ABC, abstractmethod

from rdkit import Chem
from openff.toolkit.topology.molecule import Molecule
from openff.toolkit.utils.exceptions import ToolkitUnavailableException

from ....genutils.importutils.dependencies import requires_modules
from ....genutils.decorators.functional import optional_in_place
from ....genutils.decorators.classmod import register_subclasses, register_abstract_class_attrs


def has_partial_charges(mol : Union[Molecule, Chem.Mol]) -> bool:
    '''Check if a molecular representation (either a OpenFF Molecule or and RDKit Mol) has partial charges assigned'''
    if isinstance(mol, Molecule):
        return (mol.partial_charges is not None)
    elif isinstance(mol, Chem.Mol): # NOTE : would work just as well with just an "if" here, but elif communicates intent better
        return bool(mol.HasProp('atom.dprop.PartialCharge')) # RDKit returns as int 0 or 1 instead of bool for some reason
    else:
        raise TypeError(f'Cannot check partial charge status of object of type {mol.__class__.__name__}')
    
    
# ABSTRACT AND CONCRETE CLASSES FOR CHARGING MOLECULES
@register_subclasses(key_attr='CHARGING_METHOD')
@register_abstract_class_attrs('CHARGING_METHOD')
class MolCharger(ABC):
    '''Base interface for defining various methods of generating and storing atomic partial charges'''
    @abstractmethod
    @optional_in_place
    def _charge_molecule(self, uncharged_mol : Molecule) -> None: 
        '''Method for assigning molecular partial charges - concrete implementation in child classes'''
        pass

    @optional_in_place
    def charge_molecule(self, uncharged_mol : Molecule) -> None:
        '''Wraps charge method call with logging'''
        LOGGER.info(f'Assigning partial charges via the "{self.CHARGING_METHOD}" method')
        self._charge_molecule(uncharged_mol, in_place=True) # must be called in-place for external optional_in_place wrapper to work as expected
        uncharged_mol.properties['charge_method'] = self.CHARGING_METHOD # record method within Molecule for reference
        LOGGER.info(f'Successfully assigned "{self.CHARGING_METHOD}" charges')

# CONCRETE IMPLEMENTATIONS OF DIFFERENT CHARGING METHODS
class ABE10Charger(MolCharger, CHARGING_METHOD= 'AM1-BCC-ELF10'):
    '''Charger class for AM1-BCC-ELF10 exact charging'''
    @requires_modules('openeye.oechem', 'openeye.oeomega', missing_module_error=ToolkitUnavailableException) # for whatever weird reason the toplevel openeye package has no module spec, so just checking "openeye" isn't enough
    @optional_in_place
    def _charge_molecule(self, uncharged_mol : Molecule) -> None:
        from openff.toolkit.utils.openeye_wrapper import OpenEyeToolkitWrapper
        
        uncharged_mol.assign_partial_charges(
            partial_charge_method='am1bccelf10',
            toolkit_registry=OpenEyeToolkitWrapper(), # instance init will raise exception if license or OpenEye packages are missing
        ) # TODO : find decent alternative if OpenEye license is missing (AmberTools doesn't do ELF10 and doesn't work on Windows)

class EspalomaCharger(MolCharger, CHARGING_METHOD='Espaloma-AM1-BCC'):
    '''Charger class for EspalomaCharge charging'''
    @requires_modules('espaloma_charge', missing_module_error=ToolkitUnavailableException)
    @optional_in_place
    def _charge_molecule(self, uncharged_mol : Molecule) -> None:
        from espaloma_charge.openff_wrapper import EspalomaChargeToolkitWrapper
        
        uncharged_mol.assign_partial_charges(
            partial_charge_method='espaloma-am1bcc', # this is actually the ONLY charge method the EspalomaChargeToolkitWrapper supports
            toolkit_registry=EspalomaChargeToolkitWrapper(),
        )

class NAGLCharger(MolCharger, CHARGING_METHOD='NAGL'):
    '''Charger class for NAGL charging'''
    @requires_modules('openff.nagl', 'openff.nagl_models', missing_module_error=ToolkitUnavailableException)
    @optional_in_place
    def _charge_molecule(self, uncharged_mol : Molecule) -> None:
        from openff.toolkit.utils.nagl_wrapper import NAGLToolkitWrapper
        
        uncharged_mol.assign_partial_charges(
            partial_charge_method='openff-gnn-am1bcc-0.1.0-rc.3.pt', # 'openff-gnn-am1bcc-0.1.0-rc.2.pt',
            toolkit_registry=NAGLToolkitWrapper(),
        )



'''Defining units which, for one reason or another, are not defined in Pint or OpenMM units'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import sys
_MODULE = sys.modules[__name__]

from math import pi
from openmm.unit import Unit, BaseUnit, ScaledUnit
from openmm.unit.prefix import define_prefixed_units

from openmm.unit import second_base_unit, single_item_amount_base_unit
from openmm.unit import coulomb_base_unit, ampere_base_unit, watt_base_unit
from openmm.unit import time_dimension, charge_dimension, amount_dimension
from openmm.unit import joule, erg, hartree
from openmm.unit import centimeter, gram, second
from openmm.unit import meter, candela

from scipy.constants import (
    c as speed_of_light,
    hbar as reduced_planck_const,
    electron_volt as electron_volt_in_joules,
)


# PREFIX EXPANSIONS NOT FOUND BY DEFAULT IN OPENMM UNITS
define_prefixed_units(coulomb_base_unit, module=_MODULE)
define_prefixed_units(ampere_base_unit , module=_MODULE)
define_prefixed_units(watt_base_unit   , module=_MODULE)


# UNITS FOR AMOUNTS OF ITEMS
percent_base_unit = BaseUnit(amount_dimension, 'percent', '%')
percent_base_unit.define_conversion_factor_to(single_item_amount_base_unit, 1 / 100)
percent = Unit({percent_base_unit : 1})

dozen_base_unit = BaseUnit(amount_dimension, 'dozen', 'doz')
dozen_base_unit.define_conversion_factor_to(single_item_amount_base_unit, 12)
dozen = Unit({dozen_base_unit : 1})


# PHYSICIST units - TODO : make atomic_unit_system UnitSystem subclass
## ENERGY
electronvolt_base_unit = ScaledUnit(electron_volt_in_joules, joule, 'electronvolt', 'eV')
electronvolt = eV = Unit({electronvolt_base_unit : 1.0})
define_prefixed_units(electronvolt_base_unit, module=_MODULE)

## TIME
atomic_time_base_unit = BaseUnit(time_dimension, 'atomic_time', 'au')
atomic_time_base_unit.define_conversion_factor_to(second_base_unit, reduced_planck_const / hartree.conversion_factor_to(joule))
atomic_time = Unit({atomic_time_base_unit : 1.0})

## FREQUENCY
hertz_base_unit = ScaledUnit(1.0, second**-1, 'hertz', 'Hz')
hertz = Unit({hertz_base_unit : 1.0})
define_prefixed_units(hertz_base_unit, module=_MODULE) # TODO : add conversion to radian per second (base units are incompatible though!!)


# PHOTOMETRIC UNITS
lumen_base_unit = ScaledUnit(1/(4*pi), candela, 'lumen', 'lm')
lumen = lumens = Unit({lumen_base_unit : 1})
# deliberately omitted prefix registration for now

lux_base_unit = ScaledUnit(1.0, lumen / meter**2, 'lux', 'lx')
lux = Unit({lux_base_unit : 1.0})
# deliberately omitted prefix registration for now


# CGS units
## VISCOSITY
poise_base_unit = ScaledUnit(1.0, gram/(centimeter*second), 'poise', 'P')
poise = Unit({poise_base_unit : 1.0})
define_prefixed_units(poise_base_unit, module=_MODULE)

## CHARGE
statcoulomb_base_unit = BaseUnit(charge_dimension, 'statcoulomb', 'statC') # TOSELF : evidently, this correspondence isn;t always exact
statcoulomb_base_unit.define_conversion_factor_to(coulomb_base_unit, 1/(10*speed_of_light)) # derived from the speed of light
statcoulomb = Unit({statcoulomb_base_unit : 1.0})
# deliberately omitted prefix registration for now

# ELECTRICAL POTENTIAL
statvolt_base_unit = ScaledUnit(1.0, erg / statcoulomb, 'statvolt', 'statV')
statvolt = Unit({statvolt_base_unit : 1.0})
# deliberately omitted prefix registration for now


'''Tools for simplifying transfer and modification of wrapped function type signatures'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from inspect import Parameter, Signature

POSITIONAL_PARAMETER_TYPES = [
    Parameter.POSITIONAL_ONLY,
    Parameter.POSITIONAL_OR_KEYWORD,
    Parameter.VAR_POSITIONAL
]

def get_index_after_positionals(sig : Signature) -> int:
    '''Get the first Parameter index which follows all positional Parameters'''
    for i, param in enumerate(sig.parameters.values()):
        if param.kind not in POSITIONAL_PARAMETER_TYPES:
            return i + 1 # return index immediately after first non-positional argument
        else:
            return len(sig.parameters)

def insert_parameter_at_index(sig : Signature, new_param : Parameter, index : int) -> Signature:
    '''Insert a new Parameter into a Signature at a given position'''
    params = list(sig.parameters.values())
    params.insert(index, new_param)

    return sig.replace(parameters=params)

def modify_param_annotation_by_index(sig : Signature, index : int, new_type : type) -> Signature:
    '''Returns a copy of a Signature with the type annotation of the Parameter at a given index swapped out'''
    params = list(sig.parameters.values())
    old_param = params[index] # will raise IndexError if position is given; IT IS UP TO THE CALLER TO ENSURE THIS IS CORRECT!

    new_param = Parameter( # copy everything expect type annotation from old Parameter
        name=old_param.name,
        default=old_param.default,
        annotation=new_type,
        kind=old_param.kind
    )
    params[index] = new_param

    return sig.replace(parameters=params)

def modify_param_annotation_by_name(sig : Signature, param_name : str, new_type : type) -> Signature:
    '''Returns a copy of a Signature with the type annotation of the Parameter with a given name swapped out'''
    return modify_param_annotation_by_index(
        sig,
        index=list(sig.parameters.keys()).index(param_name), # look up location of target parameter
        new_type=new_type
    )


'''SMILES and SMARTS primitives and functions for validation'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import TypeAlias

from rdkit import Chem
from rdkit.Chem.rdchem import BondType


# VALIDATION
Smiles : TypeAlias = str # purely for improving self-documentation of functions, no benefit to static type-checkers
Smarts : TypeAlias = str # purely for improving self-documentation of functions, no benefit to static type-checkers

def is_valid_SMARTS(smarts : Smarts) -> bool:
    '''Check if SMARTS string is valid (according to RDKit)'''
    return (Chem.MolFromSmarts(smarts) is not None)

def is_valid_SMILES(smiles : Smiles) -> bool:
    '''Check if SMARTS string is valid (according to RDKit)'''
    return (Chem.MolFromSmiles(smiles) is not None)

# BOND PRIMITIVES AND RELATED OBJECTS
BOND_PRIMITIVES = '~-=#$:'
BOND_PRIMITIVES_FOR_REGEX = r'[~\-=#$:]' # any of the SMARTS bond primitive chars, with a space to differentiate single-bond hyphen for the regex range char
BOND_INITIALIZERS = {
    'SMILES' : (Chem.Bond     , Chem.BondFromSmiles),
    'SMARTS' : (Chem.QueryBond, Chem.BondFromSmarts),
}
    
for in_line_fmt, (rd_type, rd_initializer) in BOND_INITIALIZERS.items():
    in_line_fmt = in_line_fmt.upper()
    rd_prefix   = rd_type.__name__.upper()

    globals()[f'BONDTYPE_BY_BOND_{in_line_fmt}'] = type_by_bonds  = {}
    globals()[f'ORDER_BY_BOND_{in_line_fmt}'   ] = order_by_bonds = {}
    globals()[f'BOND_{in_line_fmt}_BY_BONDTYPE'] = bonds_by_type  = {}
    globals()[f'BOND_{in_line_fmt}_BY_ORDER'   ] = bonds_by_order = {}
    globals()[f'RDKIT_{rd_prefix}S_BY_BONDTYPE'] = rdbonds_by_type  = {}
    globals()[f'RDKIT_{rd_prefix}S_BY_ORDER'   ] = rdbonds_by_order = {}
    
    for prim_str in BOND_PRIMITIVES:
        rd_bond = rd_initializer(prim_str)
        if (rd_bond is not None) and (type(rd_bond) == rd_type):
            bondtype, order = rd_bond.GetBondType(), rd_bond.GetBondTypeAsDouble()
            type_by_bonds[prim_str]   = bondtype
            order_by_bonds[prim_str]  = order
            bonds_by_type[bondtype]   = prim_str
            bonds_by_order[order]     = prim_str
            rdbonds_by_type[bondtype] = rd_bond
            rdbonds_by_order[order]   = rd_bond



'''Decorators for modifying other decorators'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Concatenate, Callable, ParamSpec, TypeAlias, TypeVar
from functools import update_wrapper, wraps

Params = ParamSpec('Params') # can also use to typehint *args and **kwargs
ReturnType = TypeVar('ReturnType') 
Decorator : TypeAlias = Callable[[Callable[Params, ReturnType]], Callable[Params, ReturnType]]


# META DECORATORS
def extend_to_methods(dec : Decorator) -> Decorator:
    '''Meta-decorator; modifies an existing decorator definition to be transferrable to methods with no additional code
    The modified decorator can be used interchangably to decorate both ordinary functions AND methods of classes'''
    ReturnSignature = dec.__annotations__.get('return')

    @wraps(dec, updated=()) # transfer decorator signature to decorator adapter class, without updating the __dict__ field
    class AdaptedDecorator:
        def __init__(self, funct : Callable[Params, ReturnType]) -> None:
            '''Record function'''
            self.funct = funct
            update_wrapper(self, funct) # equivalent to functools.wraps, transfers docstring, module, etc. for documentation

        def __call__(self, *args : Params.args, **kwargs : Params.kwargs) -> ReturnSignature: # TODO : fix this to reflect the decorator's return signature
            '''Apply decorator to function, then call decorated function'''
            return dec(self.funct)(*args, **kwargs)

        def __get__(self, instance : object, owner : type) -> Callable[[Concatenate[object, Params]], ReturnType]:
            '''Generate partial application with calling instance as first argument (fills in for "self")'''
            method = self.funct.__get__(instance, owner) # look up method belonging to owner class
            return dec(method) # return the decorated method

    return AdaptedDecorator


'''Thin wrappers around mdtraj-implemented property calculations'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Any, Callable, Iterable, Optional, TypeAlias, Union
from dataclasses import dataclass, field

import inspect
from itertools import combinations
from functools import wraps

import numpy as np
import pandas as pd
import mdtraj

from openmm.unit import Unit
from openmm.unit import nanometer, nanosecond, dimensionless

from ..unitutils.dimensions import hasunits
from ..genutils.typetools.parametric import T, Args, KWArgs
from ..genutils.decorators.signatures import modify_param_annotation_by_index


# ENCAPSULATION FOR PROPERTY CALCULATIONS
@dataclass
class PropCalc:
    '''For encapsulating property calculation methods, labels, and units'''
    calc   : Callable
    name   : str
    abbr   : str
    unit   : Unit
    kwargs : dict = field(default_factory=dict)

    @property
    def label(self) -> str:
        return f'{self.name} ({self.abbr}, {self.unit.get_symbol()})'

    def compute(self, traj : mdtraj.Trajectory) -> Any:
        '''Apply method to a trajectory'''
        return self.calc(traj, **self.kwargs)
    

DEFAULT_PROPS = [ # the base properties of interest for the 2023 monomer spec study
    PropCalc(calc=mdtraj.compute_rg                , name='Radius of Gyration'             , abbr='Rg'  , unit=nanometer),
    PropCalc(calc=mdtraj.shrake_rupley             , name='Solvent Accessible Surface Area', abbr='SASA', unit=nanometer**2, kwargs={'mode' : 'residue'}),
    PropCalc(calc=mdtraj.relative_shape_antisotropy, name='Relative Shape Anisotropy'      , abbr='K2'  , unit=dimensionless)    
]


# TRAJECTORY TOPOLOGY SELECTION
PairDict : TypeAlias = dict[str, Iterable[tuple[int, int]]] # a dictionary with key-labelled arrays of atom index pairs

def allow_traj_as_top(funct : Callable[[mdtraj.Topology, Args, KWArgs], T]) -> Callable[[mdtraj.Topology, Args, KWArgs], T]:
    '''Decorator which allows functions which expect an mdtraj.Topology to also accept an mdtraj.Trajectory'''
    old_sig = inspect.signature(funct) # lookup old type signature
    
    @wraps(funct) # copy over original function signature
    def mdtop_wrapper(mdtop : Union[mdtraj.Topology, mdtraj.Trajectory], *args : Args, **kwargs : KWArgs) -> T:
        if isinstance(mdtop, mdtraj.Trajectory):
            mdtop = mdtop.topology # extract Topology if Trajectory is passed
        return funct(mdtop, *args, **kwargs)
        
    # MODIFY SIGNATURE OF PATH-LIKE FIRST ARGUMENT TO MATCH NEW TYPE FLEXIBILITY
    mdtop_wrapper.__signature__ = modify_param_annotation_by_index(
        old_sig,
        index=0, # modify signature of first argument to reflect new type flexibility
        new_type=Union[mdtraj.Topology, mdtraj.Trajectory]
    )
    return mdtop_wrapper


@allow_traj_as_top
def unique_elem_types(mdtop : mdtraj.Topology) -> set[str]:
    '''Return all unique atom symbols in an mdtraj Topology'''
    return set(atom.element.symbol for atom in mdtop.atoms)

@allow_traj_as_top
def atom_pairs_by_element(mdtop : mdtraj.Topology) -> PairDict:
    '''Returns a pair dict of atom IDs by each possible duo of distinct elements'''
    return {
        f'{elem1}-{elem2}' : mdtop.select_pairs(f'element {elem1}', f'element {elem2}')
            for (elem1, elem2) in combinations(unique_elem_types(mdtop), 2) # every possible choice of 2 distinct atom types
    }


# PROPERTY CALCULATIONS TO DATAFRAMES
def acquire_rdfs(traj : mdtraj.Trajectory, pair_dict : Optional[PairDict]=None, min_rad : float=0.0, max_rad : float=1.0, rad_unit : Unit=nanometer) -> pd.DataFrame:
    '''Takes a Trajectory and produces a DataFrame for all possible pairwise Radial Distribution Functions,
    along with the radii sampled up to the specified maximum radius (must have units!)'''
    rad_range = np.array([min_rad, max_rad]) * rad_unit
    if pair_dict is None:
        pair_dict = atom_pairs_by_element(traj)

    out_dframe = pd.DataFrame()
    for label, atom_id_pairs in pair_dict.items():
        rad_points, rdf = mdtraj.compute_rdf(traj, pairs=atom_id_pairs, r_range=rad_range._value)
        out_dframe[f'Radius ({rad_unit})'] = rad_points # NOTE : radii range will be same for all pairs, so overwrite beyond first RDF is OK
        out_dframe[f'g(r) ({label})'] = rdf 

    return out_dframe

def acquire_time_props(traj : mdtraj.Trajectory, time_points : np.ndarray, time_unit : Unit=nanosecond, properties : list[PropCalc]=DEFAULT_PROPS) -> pd.DataFrame:
    '''Compute and plot a battery of labelled and unit-ed properties over a given trajectory'''
    out_dframe = pd.DataFrame()

    if hasunits(time_points):
        time_points = time_points.in_units_of(time_unit) # convert to chosen time units - NOTE : must be done BEFORE extracting unit string
        unit_str = f' ({time_points.unit.get_symbol()})'
    else:
        unit_str = ''
    out_dframe[f'Sample Time{unit_str}'] = time_points

    for prop in properties:
        time_series = prop.compute(traj) 
        if (time_series.ndim > 1):
            time_series = time_series.sum(axis=-1) # sum over atoms / residue if multiple series are given (particular to SASA calculation)

        out_dframe[prop.label] = time_series * prop.unit

    return out_dframe

## DataFrame splitting for plotting 
def _dframe_splitter_factory(regex : str):
    '''Factory for creating splitting methods that choose an x-data column and remianinig y-data columns from a dataframe
    based on a regular expression applied to the name of the cdata columns'''
    def dframe_to_plot_data(dframe : pd.DataFrame) -> tuple[pd.DataFrame, pd.DataFrame]:
        '''Takes a DataFrame populated with RDF data and unpacks it into x and y data and labels for plotting and calculations'''
        x_data = dframe.filter(regex=regex)
        y_data = dframe[[label for label in dframe.columns if label != x_data.columns[0]]] # index props with all non-time point columns

        return x_data, y_data
    return dframe_to_plot_data

rdfs_to_plot_data   = _dframe_splitter_factory(regex='Radius')
props_to_plot_data  = _dframe_splitter_factory(regex='Time')
states_to_plot_data = _dframe_splitter_factory(regex=r'\ATime \(') # ensures that 'Elapsed Time' and 'Time Remaining' are not caught by regex as x-data (need to have only 1 column)


'''For calculating the edit distance between sequences and inspecting the edits needed to go between them'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Any, Generator, Sequence, Type, TypeVar, TypeAlias
from dataclasses import dataclass, field, replace
T = TypeVar('T')

from enum import Enum
import numpy as np

from ...typetools.numpytypes import N, M, Shape
from ...bits import int_to_bits


# REPRESENTATION CLASSES
class EditOperation(Enum): # TODO: reimplement as bitwise flags
    '''For annotating distinct kinds of sequence edits and their associated index offsets'''
    NULL         = 0 # NOTE: order of fields here intentional and CANNOT be modified!
    INSERTION    = 1 # ... this is because the bits of each index are the row and column
    DELETION     = 2 # ... offsets corresponding to that edit operation in a Wagner-Fischer matrix
    SUBSTITUTION = 3

    @property
    def bits(self) -> tuple[int, int]:
        '''Convert the integer value of the Enum field into its binary bits'''
        return tuple(int_to_bits(self.value, num_bits=2, as_list=True))
    offsets = bits # alias to make dependent code more readable

@dataclass
class EditInfo:
    '''for bundling together information about a sequence edit step'''
    edit_op  : EditOperation
    indices  : tuple[int, int]
    distance : int


# WAGNER-FISCHER MATRIX OPERATION
def compute_wf_matrix(seq1 : Sequence[T], seq2 : Sequence[T], int_type : Type=int) -> np.ndarray[Shape[N, M], int]:
    '''Compute (N+1)x(M+1) matrix of Levenshtein distances between all partial prefices of a pair of sequences
    where N and M are the lengths of the first and second sequence, respectively. Implements the Wagner-Fischer algorithm'''
    n, m = len(seq1), len(seq2)
    n_aug, m_aug = n + 1, m + 1

    # initialize matrix with all zeros apart from first row and column,
    wf_matrix = np.zeros((n_aug, m_aug), dtype=int_type)
    wf_matrix[:, 0] = np.arange(n_aug, dtype=int_type) # index along first column is same as number of deletions to get to empty sequence - NOTE: element [0, 0] overlaps here
    wf_matrix[0, :] = np.arange(m_aug, dtype=int_type) # index along first row is same as number of insertions to get from empty sequence - NOTE: element [0, 0] overlaps here

    # populate matrix by iterating on distances between sub-sequence problems
    for i, elem1 in enumerate(seq1, start=1):
        for j, elem2 in enumerate(seq2, start=1):
            wf_matrix[i, j] = 1 + np.min([
                wf_matrix[i-1, j],  # deletion at end of second sequence (after augmentation to end of first sequence)
                wf_matrix[i, j-1],  # insertion at end of second sequence
                wf_matrix[i-1, j-1] - int(elem1 == elem2) # substition of last elements between sequences (if elements are equal, then substitution costs nothing)
            ])
     # TODO: implement support for transposition weighting (i.e. Damerau-Levenshtein distance)

    return wf_matrix

def traverse_wf_matrix(wf_matrix : np.ndarray[Shape[N, M], int], begin_idxs : tuple[int, int]=(0, 0), end_idxs : tuple[int, int]=(-1, -1)) -> Generator[list[EditInfo], None, None]:
    '''Takes a Wagner-Fischer Levenshtein distance matrix and returns the indices of the minimal path through the matrix
    from the origin (i.e. empty sequences) to the '''
    assert(wf_matrix.ndim == 2)
    n_aug, m_aug = wf_matrix.shape

    if end_idxs == begin_idxs:
        yield [ # need to wrap in list for consistent typing in recursive calls
            EditInfo( # base case
                edit_op=EditOperation.NULL,
                indices=begin_idxs,
                distance=0
            )
        ]
    else:
        i, j = end_idxs
        i %= n_aug # ensure values are positive
        j %= m_aug # ensure values are positive
        curr_dist = wf_matrix[i, j]

        prev_edits = []
        for edit_op in EditOperation:
            if edit_op == EditOperation.NULL:
                continue # need to skip over to avoid RecursionError - TOSELF: would be nice to reimplement EditOperations as bitwise flags to streamline this

            di, dj = edit_op.offsets # unpack index offsets
            i_prev, j_prev = i - di, j - dj
            prev_edits.append(
                EditInfo(
                    edit_op=edit_op,
                    indices=(i_prev, j_prev),
                    distance=wf_matrix[i_prev, j_prev]
                )
            )

        min_prev_dist = min(ei.distance for ei in prev_edits)
        for edit_info in prev_edits:
            ret_edit_info = replace(edit_info, indices=(i, j)) # create a copy of the current edit info which reports the current 
            if (edit_info.edit_op == EditOperation.SUBSTITUTION) and (curr_dist == min_prev_dist):
                ret_edit_info.edit_op = EditOperation.NULL
                
            if edit_info.distance == min_prev_dist:
                for edit_steps in traverse_wf_matrix(wf_matrix, begin_idxs=begin_idxs, end_idxs=edit_info.indices): # recursive tail call through all possible predecessors
                    yield edit_steps + [ret_edit_info]

def describe_edits(seq1 : Sequence[T], seq2 : Sequence[T], int_type : Type=int, indicator : str=' -> ', delimiter : str='\n') -> Generator[str, None, None]:
    '''Describes step-by-step the insertion, deletion, or substitution operations needed to transform one sequence into another'''
    seqs = (seq1, seq2) # need to bundle for zipping later
    wf_matrix = compute_wf_matrix(seq1, seq2, int_type=int_type)

    for edits in traverse_wf_matrix(wf_matrix):
        edit_descs : list[str] = []
        for edit_info in edits:
            edit_op = edit_info.edit_op
            if edit_op == EditOperation.NULL:
                continue

            elem_edit_str = indicator.join( # TODO : add printout for how the word being modified looks after each change (notnjust which symbols are changing)
                str(seq[idx - 1] if to_show else None) # need to subtract 1 to get index of step previous to current edit
                    for idx, to_show, seq in zip(edit_info.indices, edit_op.bits, seqs)
            )

            edit_descs.append(f'{edit_info.edit_op.name}: {elem_edit_str}')
        yield delimiter.join(edit_descs)


# EDIT DISTANCES
def levenshtein_distance(seq1 : Sequence[T], seq2 : Sequence[T], int_type : Type=int) -> int:
    '''Compute the Levenshtein (edit) distance between a pair of sequences with elements of compatible type
    Denotes the minimal number of insertion, deletion, or substitution operations needed to transform either sequence into the other'''
    return compute_wf_matrix(seq1, seq2, int_type=int_type)[-1, -1]


'''Unit tests for DISCERNMENT-related functionality'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import pytest
from polymerist.genutils.importutils.pkginspect import get_file_path_within_package
from polymerist.tests import data as testdata

import json
from pathlib import Path

from polymerist.genutils.sequences.discernment.inventory import SymbolInventory
from polymerist.genutils.sequences.discernment.strategies import DISCERNMENTStrategy


# DEFINE/LOAD HARD-CODED INPUTS AND EXPECTED OUTPUTS TO A PARTICULAR DISCERNMENT PROBLEM
@pytest.fixture(scope='module')
def word() -> str:
    return 'accg'

@pytest.fixture(scope='module')
def choice_bins() -> str:
    return ('bbc','aced','bad','daea','fccce','g','abcd','fggegc')

@pytest.fixture(scope='module')
def symbol_inventory(choice_bins) -> SymbolInventory:
    return SymbolInventory.from_bins(choice_bins)

@pytest.fixture(scope='module')
def solution_path() -> Path:
    return get_file_path_within_package('correct_discernment_solution.json', testdata)

@pytest.fixture(scope='module') 
def correct_solution(solution_path) -> set[tuple[int, ...]]:
    with solution_path.open('r') as file:
        solution = set(
            tuple(indices)
                for indices in json.load(file)
        )
    return solution


@pytest.mark.parametrize("ignore_multiplicities,unique_bins,DSClass", [(False, False, DSClass) for DSClass in DISCERNMENTStrategy.__subclasses__()]) # TODO: produce solutions with unique bins AND ignored multiplicities to fully test
class TestDISCERNMENTStrategies:
    all_results : dict[str, set[int]] = {} # cache solutions to avoid tedoius recalculations
    def test_preserves_symbol_inventory(
        self,
        word : str,
        symbol_inventory : SymbolInventory,
        correct_solution : set[tuple[int, ...]],
        ignore_multiplicities : bool,
        unique_bins : bool,
        DSClass : type[DISCERNMENTStrategy],
    ) -> None:
        '''Check to ensure that all implementations of DISCERNMENT solution strageties yield the same outputs and don't modify a provided symbol inventory
        Raises failure-specific Exception if inconsistency is detected, terminates silently (no Exception, returns None) otherwise'''
        mod_sym_inv = symbol_inventory.deepcopy() # create a copy of the symbol inventory to ensure any errant modifications do not affect other tests

        method_name = DSClass.__name__
        ds_strat = DSClass()
        proposed_solution = set(
            idxs
                for idxs in ds_strat.enumerate_choice_labels(
                    word,
                    mod_sym_inv,
                    ignore_multiplicities=ignore_multiplicities,
                    unique_bins=unique_bins
                )
        )
        self.all_results[method_name] = proposed_solution # cache for comparison in later tests
        assert (mod_sym_inv == symbol_inventory), f'Algorithm {method_name} does not produce to correct enumeration of bin labels'

    def test_solution_is_correct(
        self,
        word : str,
        symbol_inventory : SymbolInventory,
        correct_solution : set[tuple[int, ...]],
        ignore_multiplicities : bool,
        unique_bins : bool,
        DSClass : type[DISCERNMENTStrategy],
    ) -> None:
        method_name = DSClass.__name__
        assert (self.all_results[method_name] == correct_solution), f'Algorithm {method_name} does not return symbol inventory to original state after completion'

    def test_solution_strategies_are_consistent(
        self,
        word : str,
        symbol_inventory : SymbolInventory,
        correct_solution : set[tuple[int, ...]],
        ignore_multiplicities : bool,
        unique_bins : bool,
        DSClass : type[DISCERNMENTStrategy],
    ) -> None:
        method_name = DSClass.__name__
        proposed_solution = self.all_results[method_name]

        for other_method_name, other_solution in self.all_results.items(): # check against all other methods PRIOR TO INSERTION (minimal number of checks guaranteed to verify all pairwise checks)
            # check both symmetric differences to make sure no solution sequences are unique to either method 
            assert (proposed_solution - other_solution == set()) and (other_solution - proposed_solution == set()), f'Algorithms {method_name} and {other_method_name} produce inconsistent solutions'



'''Unit tests for `pkginspect` package`'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from types import ModuleType

import pytest
from pathlib import Path
import math, json # use these as test cases, since they are pretty stable in stdlib

from polymerist import polymerist # this is a dummy toplevel module, and NOt the entire polymerist package
from polymerist import genutils
from polymerist.genutils import importutils
from polymerist.genutils.importutils import pkginspect
from polymerist import tests



# TABULATED EXPECTED TESTS OUTPUTS
non_module_types = [ # types that are obviously not modules OR packages, and which should fail
    bool, int, float, complex, tuple, list, dict, set, # str, Path # str and Path need to be tested separately
]

are_modules = [
    ('--not_a_module--', False), # deliberately weird to ensure this never accidentally clashes with a legit module name
    (math, True),
    ('math', True), # test that the string -> module resolver also works as intended
    (json, True),
    ('json', True),
    (json.decoder, True),
    ('json.decoder', True),
    (polymerist, True),
    ('polymerist.polymerist', True),
    (genutils, True),
    ('polymerist.genutils', True),
]

are_packages = [
    ('--not_a_package--', False), # deliberately weird to ensure this never accidentally clashes with a legit module name
    (math, False),
    ('math', False), # test that the string -> module resolver also works as intended
    (json, True),
    ('json', True),
    (json.decoder, False),
    ('json.decoder', False),
    (polymerist, False),
    ('polymerist.polymerist', False),
    (genutils, True),
    ('polymerist.genutils', True),
]


# MODULE AND PACKAGE PERCEPTION
@pytest.mark.parametrize('module, expected_output', are_modules)
def test_is_module(module : ModuleType, expected_output : bool) -> None:
    '''See if Python module perception behaves as expected'''
    assert pkginspect.is_module(module) == expected_output

@pytest.mark.parametrize('non_module_type', non_module_types)
def test_is_module_fail_on_invalid_types(non_module_type : type) -> None:
    '''check that module perception fails on invalid inputs'''
    with pytest.raises(AttributeError) as err_info:
        instance = non_module_type() # create a default instance
        _ = pkginspect.is_module(instance)

@pytest.mark.parametrize('module, expected_output', are_packages)
def test_is_package(module : ModuleType, expected_output : bool) -> None:
    '''See if Python package perception behaves as expected'''
    assert pkginspect.is_package(module) == expected_output

@pytest.mark.parametrize('non_module_type', non_module_types) # NOTE: these args are in fact deliberately NOT renamed to ".*package" from ".*module"
def test_is_module_fail_on_invalid_types(non_module_type : type) -> None:
    '''check that package perception fails on invalid inputs'''
    with pytest.raises(AttributeError) as err_info:
        instance = non_module_type() # create a default instance
        _ = pkginspect.is_package(instance)

# FETCHING DATA FROM PACKAGES
@pytest.mark.parametrize(
    'rel_path, module',
    [
        ('data', tests),
        ('data/sample.dat', tests),
        pytest.param('daata/simple.dat', tests, marks=pytest.mark.xfail(raises=ValueError, reason="This isn't a real file", strict=True)),
        ('pkginspect.py', importutils),
        pytest.param('fake/whatever.txt', pkginspect, marks=pytest.mark.xfail(raises=TypeError, reason="Module is not a package and therefore cannot contain resources", strict=True)),
    ]
)
def test_get_resource_path(rel_path : str, module : ModuleType) -> None:
    '''Test fetching a resource (i.e. file OR dir) from a package'''
    resource_path = pkginspect.get_resource_path_within_package(rel_path, module)
    assert isinstance(resource_path, Path)

@pytest.mark.parametrize(
    'rel_path, module',
    [
        pytest.param('data', tests, marks=pytest.mark.xfail(raises=FileNotFoundError, reason="This is a directory, NOT a file", strict=True)),
        ('data/sample.dat', tests),
        pytest.param('daata/simple.dat', tests, marks=pytest.mark.xfail(raises=ValueError, reason="This isn't a real file", strict=True)),
        ('pkginspect.py', importutils),
        pytest.param('fake/whatever.txt', pkginspect, marks=pytest.mark.xfail(raises=TypeError, reason="Module is not a package and therefore cannot contain resources", strict=True)),
    ]
)
def test_get_file_path(rel_path : str, module : ModuleType) -> None:
    '''Test fetching a file (i.e. NOT a dir) from a package'''
    resource_path = pkginspect.get_file_path_within_package(rel_path, module)
    assert isinstance(resource_path, Path)

@pytest.mark.parametrize(
    'rel_path, module',
    [
        ('data', tests),
        pytest.param('data/sample.dat', tests, marks=pytest.mark.xfail(raises=NotADirectoryError, reason='This IS a real file, but not a directory', strict=True)),
        pytest.param('daata/simple.dat', tests, marks=pytest.mark.xfail(raises=ValueError, reason="This isn't a real file", strict=True)),
        pytest.param('pkginspect.py', importutils, marks=pytest.mark.xfail(raises=NotADirectoryError, reason='This IS a real file, but not a directory', strict=True)),
        pytest.param('fake/whatever.txt', pkginspect, marks=pytest.mark.xfail(raises=TypeError, reason="Module is not a package and therefore cannot contain resources", strict=True)),
    ]
)
def test_get_dir_path(rel_path : str, module : ModuleType) -> None:
    '''Test fetching a dir (i.e. NOT a file) from a package'''
    resource_path = pkginspect.get_dir_path_within_package(rel_path, module)
    assert isinstance(resource_path, Path)


'''Wrappers for printing out tree-like data structures'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Callable, Iterable, TypeVar, Union

from anytree import Node
from anytree.render import RenderTree
from anytree.render import AbstractStyle, AsciiStyle, ContStyle, ContRoundStyle, DoubleStyle

_render_style_aliases : dict[AbstractStyle, list[str]] = {
    AsciiStyle     : ['ASCII', 'Ascii', 'ascii'],
    ContStyle      : ['Cont', 'cont'],
    ContRoundStyle : ['Contround', 'contround', 'cont_round', 'cround', 'round'],
    DoubleStyle    : ['Doublestyle', 'Double', 'double', 'dub'],
}
RENDER_STYLE_MAP : dict[str, AbstractStyle] = {}
for StyleType in AbstractStyle.__subclasses__():
    RENDER_STYLE_MAP[StyleType.__name__] = StyleType
    for alias in _render_style_aliases[StyleType]: # register aliases for convenience
        RENDER_STYLE_MAP[alias] = StyleType


def treestr(root : Node, attr : str='name', style : Union[str, AbstractStyle]=ContStyle(), childiter : Callable[[tuple[Node]], Iterable[Node]]=list, maxlevel : int=None) -> str:
    '''Return a printable string representation of a tree from a root Node, reminiscent of GNU tree'''
    if isinstance(style, str):
        StyleType = RENDER_STYLE_MAP[style] # will raise KeyError if undefined style is provided
        style = StyleType()

    return RenderTree(
        node=root,
        style=style,
        childiter=childiter,
        maxlevel=maxlevel,
    ).by_attr(attr)


'''General-purpose utilities related to SMILES and SMARTS string manipulations'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from .primitives import is_valid_SMILES, is_valid_SMARTS, Smiles, Smarts


'''Backend web-scraping to (re)build SMARTS lookup table from the Daylight SMARTS official site'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from dataclasses import dataclass

import requests
from bs4 import BeautifulSoup

import pandas as pd


@dataclass(frozen=True)
class FnGroupSMARTSEntry:
    '''For encapuslating SMARTS group info from Daylight SMARTS registry'''
    category   : str
    category_desc : str

    group_type : str
    group_name : str

    SMARTS : str
    SMARTS_desc : str

DAYLIGHT_URL = 'https://www.daylight.com/dayhtml_tutorials/languages/smarts/smarts_examples.html'

def scrape_SMARTS(url : str=DAYLIGHT_URL) -> pd.DataFrame:
    '''Scrape SMARTS strings and accompanying descriptions and categories off of Daylight SMARTS official site'''
    soup = BeautifulSoup(requests.get(url).content, 'html.parser')

    entries = set()
    for desc_list in soup.find_all('dl'):
        category = desc_list.find_previous('a')['name']
        category_desc = desc_list.find_previous('h2').text

        group_type = desc_list.find_previous('h3').text
        for desc_term in desc_list.find_all('dt')[::-1]: # deal with annoying nesting by iterating over innermost terms first (i.e. in reverse)
            term = desc_term.extract() # remove tag from tree to prevent it from occurring in duplicate in higher terms
            text = list(term.stripped_strings)
            
            if len(text) == 3:
                group_name, SMARTS, SMARTS_desc = term.stripped_strings
            elif len(text) == 2: # raised when attempting to unpack with wrong number of args when no description was provided
                group_name, SMARTS, SMARTS_desc = *term.stripped_strings, ''
            else:
                pass
                # print(len(text), any('example' in w.lower() for w in text), text)

            entry = FnGroupSMARTSEntry(category, category_desc, group_type, group_name, SMARTS, SMARTS_desc)
            entries.add(entry)

    return pd.DataFrame.from_records(
            entry.__dict__
                for entry in entries
    )


'''Tools for editing and manipulating images, and image colors, sizes, and representations'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Union

import numpy as np
from io import BytesIO

import PIL
from PIL.Image import Image


# CONVERSION FUNCTIONS
def img_from_bytes(img_bytes : bytearray) -> Image:
    '''Load an image from a bytestream'''
    return PIL.Image.open(BytesIO(img_bytes))

def img_to_array(image : Image, encoding : str='RGB') -> np.ndarray:
    '''Convert an image to a numpy array'''
    return np.asarray(image.convert(encoding))

# BOUNDING BOXES AND BACKGROUND REMOVAL
def get_axis_bounds(image : Image, bg_color : Union[int, tuple[int]]) -> tuple[tuple[int, int], tuple[int, int]]:
    '''Takes an image and returns a pair of tuples containing the min and max non-background
    coordinates along the x- and y-axis (respectively) for some chosen background color'''
    y, x, ch = np.where((np.asarray(image) != bg_color)) # TODO : generalize this to work for greyscale (too many dimensions)
    x_bounds = (x.min(), x.max())
    y_bounds = (y.min(), y.max())

    return x_bounds, y_bounds

def get_tight_bbox(image : Image, bg_color : Union[int, tuple[int]]) -> tuple[int, ...]:
    '''Takes an image and returns a 4-tuple of the coordinates of the tight bounding box based on a choice of background color'''
    (x_min, x_max), (y_min, y_max) = get_axis_bounds(image, bg_color=bg_color)
    tight_bbox = (x_min, y_min, x_max, y_max)

    return tight_bbox

def crop_borders(image : Image, bg_color : Union[int, tuple[int]]) -> Image:
    '''Takes an image and returns an image with all extraneous borders of a particular background color cropped off'''
    return image.crop(get_tight_bbox(image, bg_color=bg_color)) 


'''Specification for representing, defining, and characterizing selective intermolecular bond placeholders ("ports")'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import ClassVar, Generator, Optional
from dataclasses import dataclass, field

from rdkit import Chem
from rdkit.Chem.rdchem import Atom, Bond, Mol

from ...genutils.iteration import iter_len
from ...genutils.containers import UnorderedRegistry


# PORT DEFINITIONS
PORT_SCHEMA = '[#0D1]~[!#0]' # an atom of null type bonded (with exact one of any type of bond) to any non-null atom
PORT_QUERY  = Chem.MolFromSmarts(PORT_SCHEMA) # prototype molecule for a port

class MolPortError(Exception):
    '''Raised when port-related errors as encountered'''
    pass

def is_linker(rdatom : Atom) -> bool:
    '''Indicate whether an atom is a linker (intermonomer "*" type atom)'''
    return rdatom.GetAtomicNum() == 0

def get_num_linkers(rdmol : Mol) -> int:
    '''Count how many wild-type inter-molecule linker atoms are in a Mol'''
    return sum(
        is_linker(atom)
            for atom in rdmol.GetAtoms()
    )

@dataclass(frozen=True)
class Port:
    '''Class for encapsulating the components of a "port" bonding site (linker-bond-bridgehead)'''
    linker     : Atom
    bond       : Bond
    bridgehead : Atom

    bondable_flavors : ClassVar[UnorderedRegistry] = field(default=UnorderedRegistry((0, 0))) # by default, only two (0, 0)-flavor ("unlabelled") ports are bondable

    @property
    def flavor(self) -> int:
        '''Return the flavor of the port'''
        return self.linker.GetIsotope()
    
    def __eq__(self, other : 'Port') -> bool:
        if not isinstance(other, Port):
            raise TypeError(f'Cannot compare {self.__class__.__name__} to object of type {other.__class__.__name__}')
        
        return (
            self.linker.Match(other.linker) # must use explicit "Match" call to correctly compare QueryAtoms and QueryBonds (may be distinct but equivalent objects)
            and self.bond.Match(other.bond)
            and self.bridgehead.Match(other.bridgehead)
        )
    
    @staticmethod
    def are_bondable(port_1 : 'Port', port_2 : 'Port') -> bool:
        '''Determine if two port atoms can be combined into a new bond'''
        return (
            ((port_1.flavor, port_2.flavor) in Port.bondable_flavors)    # 1) port flavors are one of the registered pairs
            and port_1.bond.GetBondType() == port_2.bond.GetBondType()   # 2) the ports agree on the new bond order
            and port_1.bridgehead.GetIdx() != port_2.bridgehead.GetIdx() # 3) the two ports aren't connected to the same atom
            # and not port_1.bridgehead.Match(port_2.bridgehead) # TODO : workshop this comparison for query atoms TOSELF : this match is too general (matches by query, not identity)
        )
    
    def __hash__(self) -> int: # necessary to variation in seemingly identical RDKit atoms/bonds 
        '''Distills down unique info about a Port. Simplifies equality check and allows Ports to participate in sets as expected'''
        return hash((self.flavor, self.linker.GetIdx(), self.bridgehead.GetIdx(), self.bond.GetBondTypeAsDouble())) 
    
    def matches_flavor(self, target_flavor : Optional[int]) -> bool:
        '''Returns whether a port has a particular flavor'''
        if target_flavor is None:
            return True # always returning True if None is passed (simplifies non-specific defaults)
        
        return self.flavor == target_flavor

# PORT ENUMERATION
def get_port_ids(rdmol : Mol) -> Generator[tuple[int, int], None, None]:
    '''Get the linker and bridgehead indices of all ports found in an RDKit Mol'''
    for (linker_id, bh_id) in rdmol.GetSubstructMatches(PORT_QUERY):
        yield linker_id, bh_id # unpacked purely for self-documentation

def get_linker_ids(rdmol : Mol) -> Generator[int, None, None]:
    '''Get indices of all atoms which are ports'''
    for (linker_id, bh_id) in get_port_ids(rdmol):
        yield linker_id

def get_ports(rdmol : Mol, target_atom_id : Optional[int]=None, target_flavor : Optional[int]=None) -> Generator[Port, None, None]:
    '''Find and generate all ports in a molecule. Can optioanlly narrow scope to port whose bridegehead is a particular atom and/or whose linker has a particular flavor'''
    target_atom = None if (target_atom_id is None) else rdmol.GetAtomWithIdx(target_atom_id)

    for (linker_id, bh_id) in get_port_ids(rdmol):
        port = Port(
            linker=rdmol.GetAtomWithIdx(linker_id),
            bond=rdmol.GetBondBetweenAtoms(linker_id, bh_id),
            bridgehead=rdmol.GetAtomWithIdx(bh_id)
        )

        if port.matches_flavor(target_flavor):
            # if (target_atom is None) or port.bridgehead.Match(target_atom):# will match if no atom is given OR if an atom is given and the bridgehead matches that Atom
            if (target_atom_id is None) or (port.bridgehead.GetIdx() == target_atom_id):# will match if no atom is given OR if an atom is given and the bridgehead matches that Atom
                yield port

def get_num_ports(rdmol : Mol, target_atom_id : Optional[int]=None, target_flavor : Optional[int]=None) -> int: 
    '''Counts the number of port atoms present in a Mol'''
    return iter_len(get_ports(rdmol, target_atom_id=target_atom_id, target_flavor=target_flavor))

def get_single_port(rdmol : Mol) -> Port:
    '''Get the singular port of a Mol which contains only 1 port'''
    num_ports = get_num_ports(rdmol)
    if num_ports == 0:
        raise MolPortError("Can't return port for molecule which has no ports")
    
    if num_ports > 1:
        raise MolPortError("Molecule has multiple ports; choice of single port is ambiguous")
    
    return next(get_ports(rdmol)) # return if port count checks pass


'''For explicitly enumerating partitions of sets and multisets'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Generator, Sequence


def int_partitions(n : int, _pivot : int=1) -> Generator[tuple[int], None, None]:
    '''Enumerates all integer partitions of n'''
    yield (n,)
    for i in range(_pivot, n//2 + 1):
        for p in int_partitions(n - i, _pivot=i):
            yield (i,) + p

def multiset_partition(n : int, k : int) -> Generator[tuple[int], None, None]:
    '''Enumerates all multisets of k integers whose sum is n (respects order and includes 0s)'''
    if k == 0:
        yield ()
    elif k == 1:
        yield (n,)
    else:
        for i in range(0, n + 1):
            for subpart in multiset_partition(n - i , k - 1):
                yield (i,) + subpart

def make_change_greedy(n : int, denoms : Sequence[int]) -> dict[int, int]:
    '''Greedy algorithm for making change for a given total'''
    # precheck for uniqueness and integrality of denominations
    seen = set()
    for den in denoms:
        if den in seen:
            raise ValueError('No duplicate denominations allowed')
        elif not isinstance(den, int):
            raise TypeError('All denominations must be integers')
        else:
            seen.add(den)

    # actual change-making happens here
    multips = {}
    for den in sorted(denoms, reverse=True):
        d, n = divmod(n, den)
        multips[den] = d

    if n != 0:
        raise ValueError(f'Could not successfully make complete change ({n} left over)')
    return multips


'''Tools for generating spec-compliant monomer substructure
Code by Connor Davel, from https://github.com/openforcefield/polymer_examples/blob/main/monomer_generation/substructure_generator.py'''

__author__ = 'Connor Davel'

import itertools
from collections import defaultdict

import json
import sys, getopt
from copy import deepcopy

from rdkit import Chem


# CUSTOM EXCEPTIONS
class InvalidRDKitMol(Exception):
    # raised during validation and error checking to catch invalid inputted/generated monomers
    def __init__(self, message, smarts=""):
        self.message = message
        self.smarts = smarts
        super().__init__(self.message)
    
    def __str__(self):
        if self.smarts == "":
            return f"rdmol found to be invalid: {self.message}"
        else:
            return f"rdmol with smarts \"{self.smarts}\" found to be invalid: {self.message}"

class InvalidMonomer(InvalidRDKitMol):
    pass

class InvalidCap(InvalidRDKitMol):
    pass 
 
 # CONVERTERS
class Monomer:
    # utility class to store monomer info
    def __init__(self, name):
        self.name = name
        self.smarts = ""
        self.caps = []

class SubstructureGenerator:
    '''Utility class for curating molecule from various representations and converting them to spec-compliant SMARTS'''
    def __init__(self):
        self.monomers = defaultdict()

    #MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM
    #                      READERS
    #WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW

    def add_monomer_as_smarts_fragment(self, smarts, name, cap_ids=[], replace_map_nums=True, add_caps_from_discarded_ids=True):
        rdmol = Chem.rdmolfiles.MolFromSmarts(smarts)
        # add map numbers to any atoms 
        if replace_map_nums:
            for atom in rdmol.GetAtoms():
                atom.SetAtomMapNum(atom.GetIdx() + 1)
        else:
            self.check_map_nums()

        if cap_ids == []:
            return -1
        
        new_cap_ids = []
        for idx in cap_ids:
            if idx < 0:
                idx = rdmol.GetNumAtoms() + idx
            new_cap_ids.append(idx)
        cap_ids = new_cap_ids
        
        body = [] # opposte of caps list
        for atom in rdmol.GetAtoms():
            idx = atom.GetIdx()
            if idx not in cap_ids:
                body.append(idx)

        # caps are found by finding all connected fragments with ids in "caps"
        cap_fragments = []
        if add_caps_from_discarded_ids:
            for atom in rdmol.GetAtoms():
                id = atom.GetIdx()
                if id not in cap_ids:
                    continue
                # if neighbor in existing fragment, add id to that fragment
                connected_fragments = []
                for neighbor in atom.GetNeighbors():
                    n_id = neighbor.GetIdx()
                    for fragment_id, cap_frag in enumerate(cap_fragments):
                        if n_id in cap_frag:
                            connected_fragments.append(fragment_id)
                
                if len(connected_fragments) > 0:
                    new_fragment = {id}
                    for frag_id in sorted(connected_fragments, reverse=True):
                        frag = cap_fragments.pop(frag_id)
                        new_fragment = new_fragment | frag
                    cap_fragments.append(new_fragment)
                else:
                    cap_fragments.append(set([id]))

        # find where each monomer connects to the cap, and replace with wildtype atoms 
        # but with the same bond order 
        mid_rdmol = deepcopy(rdmol)
        body_with_connections = deepcopy(body)
        for atom in mid_rdmol.GetAtoms():
            if atom.GetIdx() in cap_ids:
                continue
            # search atom neighbors for connections to capped atoms
            for neighbor in atom.GetNeighbors():
                if neighbor.GetIdx() in cap_ids:
                    neighbor.SetAtomicNum(0)
                    neighbor.SetQuery(Chem.AtomFromSmarts("[*]")) # must reset query for wildtype atom to be printed
                    body_with_connections.append(neighbor.GetIdx())

        monomer_smarts = Chem.MolFragmentToSmarts(mid_rdmol, atomsToUse=body_with_connections)
        self.add_monomer(name, monomer_smarts)
        # finally, for each connected cap, find where the cap connects to the monomer
        for cap_fragment in cap_fragments:
            cap_ids_to_use = list(cap_fragment)
            cap_rdmol = deepcopy(rdmol)
            for atom in cap_rdmol.GetAtoms():
                if atom.GetIdx() not in cap_fragment:
                    continue
                # search atom neighbors for connections to monomer_atom
                for neighbor in atom.GetNeighbors():
                    if neighbor.GetIdx() in body: # if part of main body of monomer
                        neighbor.SetQuery(Chem.AtomFromSmarts("[*]")) # must reset query for wildtype atom to be printed
                        # neighbor should retain its original atomMapNum
                        cap_ids_to_use.append(neighbor.GetIdx())
            cap_smarts = Chem.MolFragmentToSmarts(cap_rdmol, atomsToUse = cap_ids_to_use)
            self.add_monomer_cap(name, cap_smarts)
            
    def add_monomer_as_sdf_fragment(self, sdf_file, name, caps=[], add_caps_from_discarded_ids=True):
        # identical in function to `self.add_monomer_as_smarts_fragment`, but reads monomer 
        # from an sdf file instead of a smarts. Can handle sdf with multiple molecules, but 
        # must specify a list of names or a single name string that will be enumerated from 1,2,3,...
        suppl = Chem.SDMolSupplier(sdf_file, removeHs=False)
        if isinstance(name, list): # a list of names
            name_index = 0
            for rdmol in suppl:
                try:
                    name = name[name_index]
                except IndexError:
                    print("not enough names for SDFFile, exiting")
                    return
                name_index += 1
                smarts = Chem.rdmolfiles.MolToSmarts(rdmol)
                self.add_monomer_as_smarts_fragment(smarts, name, add_caps_from_discarded_ids=add_caps_from_discarded_ids,  caps=[])
        elif isinstance(name, str): # a string that can be enumerated <str> + <name_index>
            name_index = 1
            for rdmol in suppl:
                new_name = name + str(name_index)
                name_index += 1 
                smarts = Chem.rdmolfiles.MolToSmarts(rdmol)
                print(smarts)
                self.add_monomer_as_smarts_fragment(smarts, new_name, caps=[], add_caps_from_discarded_ids=add_caps_from_discarded_ids)
        else:
            print(f"cannot handle name(s) of type: {type(name)}")

    def read_monomer_info_dict(self, monomer_dict):
        # reads monomer_dict of the same format as that returns from `self.get_monomer_info_dict`
        if "monomers" not in monomer_dict.keys():
            print(f"no \"monomers\" field present in monomer_dict out of the given keys ({monomer_dict.keys()})")
        monomers = monomer_dict['monomers']
        for name, monomer_smarts in monomers.items():
            self.add_monomer(name, monomer_smarts)
        if "caps" in monomer_dict.keys():
            for name, caps_list in monomer_dict["caps"].items():
                for cap in caps_list:
                    self.add_monomer_cap(name, cap)
        # check for miscellaneous keys
        for key in monomer_dict.keys():
            if key not in ["monomers", "caps"]:
                print(f"malformated input: key \"{key}\" is not supported, skipped")
    
    def input_monomer_info_json(self, file_name):
        with open(file_name, "r") as file:
            data = json.load(file)
            self.read_monomer_info_dict(data)

    def add_monomer(self, name, smarts):
        if name in self.monomers.keys():
            print("monomer name already exists")
            return
        monomer = Monomer(name)
        monomer.caps = []
        monomer.smarts = smarts

        self.monomers[name] = monomer
        pass

    def add_monomer_cap(self, name, smarts):
        '''
        Adds cap to an already existing monomer 
        name: name of existing monomer
        smarts: smarts of the cap to add to the monomer. Wildtype atoms ("[*]")
                represent where the cap connects to the monomer, which must contain
                the atomMapNum of the monomer atom to connect to. MUST CONTAIN EXPLICIT
                HYDROGENS
                  ex) for [*]-[C:1](-[H])(-[H])-[C:2](-[H])(-[H])-[*], a valid cap would be
                  represented as [*:1]-[H] (note how [*:1] corresponds to [C:1])
                  or maybe [*:2]-[C](-[H])(-[H])(-[H])(note how [*:2] corresponds to [C:2])
        '''
        if name not in self.monomers.keys():
            print("monomer name does not exist")
            return
        monomer = self.monomers[name]
        # TODO: add some validation here for edge cases
        monomer.caps.append(smarts)
        pass

    #MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM
    #                        WRITERS
    #WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
    def _get_maximum_map_num(self, rdmol):
        return max([atom.GetAtomMapNum() for atom in rdmol.GetAtoms()])
    
    def _fill_out_query(self, query_input):
        rdmol = Chem.MolFromSmarts(query_input.replace('&', ''))
        current_map_num = 1 + self._get_maximum_map_num(rdmol)
        for atom in rdmol.GetAtoms():
            if atom.GetAtomMapNum() == 0:
                atom.SetAtomMapNum(current_map_num)
            if atom.GetAtomicNum() > 0:
                a_num = atom.GetAtomicNum()
                D_num = len([0 for _ in atom.GetBonds()])
                F_num = atom.GetFormalCharge()
                query_string = f"[#{a_num}D{D_num}{F_num:+}:{atom.GetAtomMapNum()}]"
                query = Chem.AtomFromSmarts(query_string)
                atom.SetQuery(query)
            elif atom.GetAtomicNum() == 0:
                query_string = f"[*:{atom.GetAtomMapNum()}]"
                query = Chem.AtomFromSmarts(query_string)
                atom.SetQuery(query)
            else:
                raise Exception
            current_map_num += 1

        # fill bond queries to explicity specific singles and aromatics (these can often be omitted)
        for bond in rdmol.GetBonds():
            if bond.GetIsAromatic():
                query = Chem.BondFromSmarts(":")
                bond.SetQuery(query)
            elif bond.GetBondType() == Chem.rdchem.BondType.SINGLE:
                query = Chem.BondFromSmarts("-")
                bond.SetQuery(query)

        smarts_string = Chem.MolToSmarts(rdmol)
        return [smarts_string.replace('&', '')] # all & are removed for rdkit to read query and read molecule info
    
    def _enumerate_substructures_with_caps(self, name, remove_complete_substructures=True):
        # for a named substructure, returns all possible combinations
        # of caps and inter-monomer bonds. 
        # monomer_smarts: smarts of the monomer, with wildtype atoms (required) repressenting 
        #                 intermonomer_bonds and any possible connection to a cap
        # monomer_caps: list of smarts with atom_mapped wildtype atoms representing a connection
        #               to a atom. The AtomMapNum of the wildtype atom must match the AtomMapNum
        #               of the atom represented by the wildtype atom
        monomer = self.monomers[name]
        if monomer.caps == []:
            return {name: monomer.smarts}

        rdmol = Chem.MolFromSmarts(monomer.smarts)
        cap_groups = []
        # the default "cap" is the intermonomer bond represented with wildtype atoms
        for atom in rdmol.GetAtoms():
            if atom.GetAtomicNum() == 0:
                atom_id = atom.GetIdx()
                rdmol_copy = deepcopy(rdmol)
                atom_copy = rdmol_copy.GetAtomWithIdx(atom_id)
                # atom_copy.SetAtomMapNum(0)
                # should only have one neighbor
                next_monomer_connection = None
                for neighbor in atom_copy.GetNeighbors():
                    if neighbor.GetAtomMapNum() > 0: # if connected to intermonomer bond
                        next_monomer_connection = neighbor
                next_monomer_connection.SetAtomicNum(0)
                atom_map_num = next_monomer_connection.GetAtomMapNum()
                next_monomer_connection.SetQuery(Chem.AtomFromSmarts("[*]"))
                next_monomer_connection.SetAtomMapNum(atom_map_num)
                ids_to_include = [atom.GetIdx(), next_monomer_connection.GetIdx()]
                cap_smarts = Chem.MolFragmentToSmarts(rdmol_copy, atomsToUse = ids_to_include)
                cap_groups.append(tuple([atom_map_num, [cap_smarts]]))
        # also add end-of-polymer caps that are specified
        for cap_smarts in monomer.caps:
            cap_rdmol = Chem.MolFromSmarts(cap_smarts)
            # find with atom on the monomer the cap bonds to
            for atom in cap_rdmol.GetAtoms():
                if atom.GetAtomicNum() == 0:
                    for map_num, cap_smarts_list in cap_groups:
                        if atom.GetAtomMapNum() == map_num:
                            cap_smarts_list.append(cap_smarts)
                    break
        # enumerate all combinations of the cap_groups
        # cap_groups = [set(l) for map_num, l in cap_groups]
        # deduplified_groups = []
        # [deduplified_groups.append(i) for i in cap_groups if i not in deduplified_groups]
        # deduplified_groups = [list(i) for i in deduplified_groups]
        cap_groups = [[(map_num, string) for string in l] for map_num, l in cap_groups]
        name_id = 0
        enumerated_substructures = dict()
        for iters in itertools.product(*cap_groups): # enumerate all combinations of cap_groups
            name_id += 1
            if name_id != 1:
                cap_group_name = name + "_TERM" + str(name_id)
            else:
                cap_group_name = name
            # first, remove all wildtype atoms
            nonwildtype_atoms = []
            for atom in rdmol.GetAtoms():
                if atom.GetAtomicNum() > 0:
                    nonwildtype_atoms.append(atom.GetIdx())
            substructure_smarts = Chem.MolFragmentToSmarts(rdmol, atomsToUse = nonwildtype_atoms)
            substructure = Chem.MolFromSmarts(substructure_smarts)
            # attach caps to the substructure (except for intermonomder bonds, which are already present)
            for attachment_atom_map_num, cap_smarts in iters:
                #TODO: better way of doing this
                # attach the cap smarts to the substructure
                cap_rdmol = Chem.MolFromSmarts(cap_smarts)
                # remove attachment point
                cap_start_atom_map_num = -1
                cap_ids_to_include = []
                inter_monomer_bond_order = None
                for atom in cap_rdmol.GetAtoms():
                    if atom.GetAtomMapNum() == attachment_atom_map_num:
                        # should have only one neighbor
                        neighbor = atom.GetNeighbors()[0]
                        cap_start_atom_map_num = neighbor.GetAtomMapNum()
                        bond = cap_rdmol.GetBondBetweenAtoms(atom.GetIdx(), neighbor.GetIdx())
                        inter_monomer_bond_order = bond.GetBondType()
                    else:
                        cap_ids_to_include.append(atom.GetIdx())

                cap_fragment = Chem.MolFragmentToSmarts(cap_rdmol, atomsToUse=cap_ids_to_include)
                cap_rdmol = Chem.MolFromSmarts(cap_fragment)
                # find where to attach the cap_fragment using atom map num
                new_substructure = Chem.CombineMols(substructure, cap_rdmol)
                # find start and end ids for the bond
                start = -1
                end = -1
                for atom in new_substructure.GetAtoms():
                    if atom.GetAtomMapNum() == cap_start_atom_map_num:
                        start = atom.GetIdx()
                    elif atom.GetAtomMapNum() == attachment_atom_map_num:
                        end = atom.GetIdx()
                editable_mol = Chem.EditableMol(new_substructure)
                editable_mol.AddBond(start, end, order = inter_monomer_bond_order)
                substructure = editable_mol.GetMol()
            if remove_complete_substructures:
                incomplete_substructure = False
                for atom in substructure.GetAtoms():
                    if atom.GetAtomicNum() == 0:
                        incomplete_substructure = True
                if not incomplete_substructure:
                    continue
            enumerated_substructures[cap_group_name] = Chem.MolToSmarts(substructure)
        return enumerated_substructures

    def get_monomer_info_dict(self, expand=True):
        if expand == True:
            monomer_dict = {"monomers": defaultdict(str)}
        else:
            monomer_dict = {"monomers": defaultdict(str), "caps": defaultdict(list)}

        for name, monomer in self.monomers.items():
            if expand == True:
                enumerated_monomers = self._enumerate_substructures_with_caps(name)
                for name, capped_monomer in enumerated_monomers.items():
                    monomer_dict["monomers"][name] = self._fill_out_query(capped_monomer)
            else:
                monomer_dict["monomers"][name] = self._fill_out_query(monomer.smarts)
                monomer_dict["caps"][name] = self._fill_out_query(monomer.caps)

        return monomer_dict 

    def output_monomer_info_json(self, file_name, expand=True):
        monomer_dict = self.get_monomer_info_dict()
        with open(file_name, "w") as file:
            json.dump(monomer_dict, file, indent=4)

    #MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM
    #                     FORMATING
    #WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
    def format_smarts_query(smarts):
        return smarts

    def validate_smarts_query(smarts):
        return False
    #MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM
    #                   ERROR HANDLING
    #WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
    def check_map_nums(self, rdmol):
        # rdmol: Smarts or Chem.Mol object
        if isinstance(rdmol, str):
            rdmol = Chem.MolToSmarts(rdmol)

        existing_map_nums = []
        for atom in rdmol.GetAtoms():
            map_num = atom.GetAtomMapNum()
            if map_num == 0:
                Exception(f"All atoms must have map numbers when replace_map_nums=False (ex: [{atom.GetAtomMapNum()}:1])")
            elif map_num in existing_map_nums:
                Exception("Atom map numbers must be unique")
            existing_map_nums.append(map_num)
        return
    
    def check_cap(self, rdmol):
        if isinstance(rdmol, str):
            rdmol = Chem.MolToSmarts(rdmol)
        return



'''Conversion tools between various programming language cases (https://en.wikipedia.org/wiki/Letter_case#Use_within_programming_languages)'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

def snake_case_to_camel_case(varname : str) -> str:
    '''Convert a name from Snake Case to Camel Case
    E.g. name_of_a_thing -> NameOfAThing'''
    return ''.join(word.capitalize() for word in varname.split('_'))

def camel_case_to_snake_case(varname : str) -> str:
    '''Convert a name from Camel Case to Snake Case
    E.g. NameOfAThing -> name_of_a_thing'''
    cap_idxs = [i for i, char in enumerate(varname) if char.isupper()]
    return '_'.join(
        varname[i_start:i_end].lower()
            for i_start, i_end in zip(cap_idxs, cap_idxs[1:]+[None])
    ) 


'''Front-facing solver facade for the DISCERNMENT Problem'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Generator, Iterable, Mapping, Sequence, Union

from .inventory import SymbolInventory, T, L
from .strategies import DISCERNMENTStrategy, DISCERNMENTStrategyStack


# TODO : add custom Exceptions to provide more detailed feedback?
class DISCERNMENTSolver():
    '''Encapsulation class for solving generalized ransom-note index enumeration problems for arbitrary words, bins, and solving algorithms'''
    def __init__(self, symbol_inventory : Union[SymbolInventory[T, L], Sequence[Iterable[T]], Mapping[L, Sequence[Iterable[T]]]], strategy : DISCERNMENTStrategy=DISCERNMENTStrategyStack()) -> None:
        if isinstance(symbol_inventory, SymbolInventory):
            self._symbol_inventory = symbol_inventory
        else:
            self._symbol_inventory = SymbolInventory.from_bins(symbol_inventory)
        self.strategy = strategy

    @property
    def symbol_inventory(self) -> SymbolInventory[T, L]:
        '''Cast symbol inventory as copy to avoid mutation during partial traversals (i.e. peek at first item)'''
        return self._symbol_inventory.deepcopy()

    def enumerate_choices(self, word : Sequence[T], ignore_multiplicities : bool=False, unique_bins : bool=False) -> Generator[L, None, None]:
        '''Enumerate all possible choices using specified solution strategy'''
        return self.strategy.enumerate_choice_labels(
            word=word,
            symbol_inventory=self.symbol_inventory,
            ignore_multiplicities=ignore_multiplicities,
            unique_bins=unique_bins,
        )

    def choice_solutions_exist(self, word : Sequence[T], ignore_multiplicities : bool=False, unique_bins : bool=False) -> bool:
        '''Precheck to see if a solution exists without attemting full enumeration'''
        if not self.symbol_inventory.contains_word(word, ignore_multiplicities=ignore_multiplicities):
            return False
        try:
            first_solution = next(self.enumerate_choices(word, ignore_multiplicities=ignore_multiplicities, unique_bins=unique_bins))
            return (first_solution != tuple()) # consider the null index sequence of the empty tuple to be a failure
        except StopIteration:
            return False
    choices_exist = choice_exists = solutions_exist = solution_exists = choice_solutions_exist # aliases for convenience
        
